/*	bench.c

	Program to benchmark the AFAPI.

	This program simply runs every individual AFAPI routine many
	times with random data derived from a seed given on the
	command line -- what many people now refer to as
	"microbenchmarking."  Rather than running the routines a fixed
	number of times, a signal is set for one second, and the
	handler for this signal simply sets a flag that allows the
	test loop to terminate.  Of course, the catch is that the loop
	should not terminate until the timer expiration is known by
	all; thus, a p_any() is used in each loop iteration to notify
	everybody of the timer expiration on PE CPROC.  You can set the
	timer interval by changing the define of SECS below.

	In experimenting with various systems and flavors of UNIX, it
	seems that the itimer values are often less precise than the
	values returned by gettimeofday().  Thus, all processors use
	that to compute the actual elapsed time, and the minimum
	locally-computed value is used (slower values may reflect
	extra delay in processing the timer calls).  The times taken
	by loop setup and an empty p_any() loop are measured first, so
	that these overheads can be subtracted out.  For each AFAPI
	routine, the adjusted time divided by the number of function
	calls completed (i.e., average time per call in microseconds)
	is output to standard out.  There is still a little overhead in
	signal handling that I haven't accounted for, but that should
	be well below the noise level for our timings.

	Ok, all the above applies to p_ routines...  how do we make
	this also work for the s_ stuff?  Well, it shouldn't be too
	bad...  we simply need to do things a little differently in
	terms of the logic for sending things, because p_any() is not
	available in the s_ world.

	1997 by Hank Dietz
*/


#define	SECS	2	/* Basic timing interval (in seconds) */


#include <stdlib.h>
#include <stdio.h>
#include <AFAPI/afapi.h>
#include <AFAPI/inline.h>

#include "rand.h"


#define	TABSIZE	1024
#define	MODTABSIZE(x)	((TABSIZE - 1) & (x))

#define	P_DECLARE(suffix) \
p_##suffix tab_##suffix[TABSIZE];
P_ALPHA
#undef	P_DECLARE
int tab_PEno[TABSIZE];		/* Unique PE numbers */
int tab_only1[TABSIZE];		/* 1 for only one PE (sender) */


#define	R_bar	tab_bar[MODTABSIZE(count)]
#define	R_1u	tab_1u[MODTABSIZE(count)]
#define	R_8	tab_8[MODTABSIZE(count)]
#define	R_8u	tab_8u[MODTABSIZE(count)]
#define	R_16	tab_16[MODTABSIZE(count)]
#define	R_16u	tab_16u[MODTABSIZE(count)]
#define	R_32	tab_32[MODTABSIZE(count)]
#define	R_32u	tab_32u[MODTABSIZE(count)]
#define	R_64	tab_64[MODTABSIZE(count)]
#define	R_64u	tab_64u[MODTABSIZE(count)]
#define	R_32f	tab_32f[MODTABSIZE(count)]
#define	R_64f	tab_64f[MODTABSIZE(count)]
#define	R_ptr	tab_ptr[MODTABSIZE(count)]
#define	R_PEno	tab_PEno[MODTABSIZE(count)]
#define	R_only1	tab_only1[MODTABSIZE(count)]


void
setup_tabs(void)
{
	register int i;

	for (i=0; i<TABSIZE; ++i) {
#define	P_DECLARE(suffix) \
		tab_##suffix[i] = p_randUniq##suffix();
P_ALPHA
#undef	P_DECLARE
		tab_PEno[i] = (p_randUniq16u() % NPROC);
		tab_only1[i] = (IPROC == (p_randSame16u() % NPROC));
	}
}


/*	The timer support stuff...
*/
#include <sys/time.h>
#include <signal.h>

volatile p_8u timeup = 0;
void
gotsig(register int dummy)
{
	++timeup;
	signal(SIGALRM, gotsig);
}
p_64f tonce = 0.0;
p_64f tperloop = 0.0;

p_64f stonce = 0.0;
p_64f stperloop = 0.0;
static struct timeval btime, etime;

void
begint(void)
{
	/* Make sure everybody is here */
	signal(SIGALRM, gotsig);
	timeup = 0;
	p_wait();

	/* Only PE CPROC does the signaling */
	if (IPROC == CPROC) {
		struct itimerval newt, oldt;

		newt.it_interval.tv_sec =
		newt.it_value.tv_sec = SECS;
		newt.it_interval.tv_usec =
		newt.it_value.tv_usec = 0;
		setitimer(ITIMER_REAL, &newt, &oldt);
	}

	/* Everybody does the timing... */
	gettimeofday(&btime, 0);

	/* Make sure everybody is here */
	timeup = 0;
	p_wait();
}

int
endt(void)
{
	/* Return 1 when timer has expired; else 0 */
	return(p_any(timeup));
}

int
endts(void)
{
	/* Return 1 when timer has expired; else 0 */
	register int t;

	s_exit();
	t = p_any(timeup);
	s_init();
	return(t);
}

p_64f
usec(register p_32u calls)
{
	/* Compute microseconds per call... */
	struct itimerval late;
	register p_64f t;

	p_wait();

	/* Everybody does the timing and we take fastest... */
	gettimeofday(&etime, 0);
	t = ((1000000.0 * (etime.tv_sec - btime.tv_sec)) +
	     (etime.tv_usec - btime.tv_usec) + 0.5);
	t -= tonce;
	t /= calls;
	t -= tperloop;
	return(p_reduceMin64f(t));
}

p_64f
susec(register p_32u calls)
{
	/* Compute microseconds per call... */
	struct itimerval late;
	register p_64f t;

	p_wait();

	/* Everybody does the timing and we take fastest... */
	gettimeofday(&etime, 0);
	t = ((1000000.0 * (etime.tv_sec - btime.tv_sec)) +
	     (etime.tv_usec - btime.tv_usec) + 0.5);
	t -= stonce;
	t /= calls;
	t -= stperloop;
	return(p_reduceMin64f(t));
}


#define	BEGINT \
{ \
	register p_32u count = 0; \
	begint(); \
	do {

#define	ENDT(name) \
		++count; \
	} while (!endt()); \
	printf(name "\t%.1f\n", usec(count)); \
}


#define	BEGINTS \
{ \
	register p_32u count = 0; \
	begint(); \
	s_init(); \
	do {

#define	ENDTS(name) \
		++count; \
	} while (!endts()); \
	s_exit(); \
	printf(name "\t%.1f\n", susec(count)); \
}


int
less64f(register const p_64f *a,
register const p_64f *b)
{
	if (*a < *b) return(-1);
	else if (*a > *b) return(1);
	else return(0);
}

void
calibrate(void)
{
	/* Calibrate the tonce and tperloop factors... */
	p_64f median[201];
	register int i;
	register p_32u count = 0;

	for (i=0; i<201; ++i) {
		begint();
		median[i] = usec(1);
	}
	qsort(&(median[0]), 201, sizeof(p_64f), 
		(int (*)(const void*,const void*)) less64f);
	tonce = median[100];
	printf("One-time overhead calibration\t%.1f\n", tonce);

	for (i=0; i<11; ++i) {
		count = 0;
		begint();
		do {
			volatile p_32u t = R_32u;
			++count;
		} while (!endt());
		median[i] = usec(count);
	}
	qsort(&(median[0]), 11, sizeof(p_64f),
		(int (*)(const void*,const void*)) less64f);
	tperloop = median[5];
	printf("Per-loop overhead calibration\t%.1f\n", tperloop);
}


void
scalibrate(void)
{
	/* Calibrate the stonce and stperloop factors... */
	p_64f median[201];
	register int i;
	register p_32u count = 0;

	for (i=0; i<201; ++i) {
		begint();
		s_init();
		s_exit();
		median[i] = susec(1);
	}
	qsort(&(median[0]), 201, sizeof(p_64f),
		(int (*)(const void*,const void*)) less64f);
	stonce = median[100];
	printf("One-time overhead s_ calibration\t%.1f\n", stonce);

	for (i=0; i<11; ++i) {
		count = 0;
		begint();
		s_init();
		do {
			volatile p_32u t;
			t = R_32u;
			++count;
		} while (!endts());
		s_exit();
		median[i] = susec(count);
	}
	qsort(&(median[0]), 11, sizeof(p_64f),
		(int (*)(const void*,const void*)) less64f);
	stperloop = median[5];
	printf("Per-loop overhead s_ calibration\t%.1f\n", stperloop);
}


void
control(void)
{
	BEGINT
	p_exit(); p_init();
	ENDT("void p_exit(void); int p_init(void)")

	BEGINT
	p_confirm();
	ENDT("int p_confirm(void)");

	putchar('\n');
}

void
primitives(void)
{
	BEGINT
	p__wait();
	ENDT("void p__wait(void)");

	BEGINT
	p_wait();
	ENDT("void p_wait(void)");

	BEGINT
	p__waitbar(R_1u);
	ENDT("p_bar p__waitbar(p_1u)");

	BEGINT
	p_waitbar(R_1u);
	ENDT("p_bar p_waitbar(p_1u)");

	BEGINT
	p__any(R_1u);
	ENDT("p_1u p__any(p_1u)");

	BEGINT
	p_any(R_1u);
	ENDT("p_1u p_any(p_1u)");

	BEGINT
	p__all(R_1u);
	ENDT("p_1u p__all(p_1u)");

	BEGINT
	p_all(R_1u);
	ENDT("p_1u p_all(p_1u)");

	putchar('\n');
}


void
communications(void)
{
#define	P_DECLARE(suffix) \
	BEGINT \
	if (R_only1) { \
		p_bcastPut##suffix(R_##suffix); \
	} else { \
		p_bcastGet##suffix(); \
	} \
	ENDT("p_" #suffix " p_bcastPut" #suffix "(p_" #suffix ") / p_" #suffix " p_bcastGet" #suffix "(void)")
P_ALPHA
#undef	P_DECLARE

	BEGINT
	p_8u dummy[TABSIZE];
	if (R_only1) {
		p_bcastPutz(&(tab_8u[0]), TABSIZE);
	} else {
		p_bcastGetz(&(dummy[0]), TABSIZE);
	}
	ENDT("void p_bcastPutz(ptr, int (1024)) / void p_bcastGetz(ptr, int (1024))")

#define	P_DECLARE(suffix) \
	BEGINT \
	p_putGet##suffix(R_##suffix, R_PEno); \
	ENDT("p_" #suffix " p_putGet" #suffix "(p_" #suffix ", int)")
P_ALPHA
#undef	P_DECLARE

	BEGINT
	p_8u dummy[TABSIZE];
	p_putGetz(&(dummy[0]), &(tab_8u[0]), TABSIZE, R_PEno);
	ENDT("void p_putGetz(ptr, ptr, int (1024), int)")

#define	P_DECLARE(suffix) \
	BEGINT \
	p_##suffix dummy[TABSIZE]; \
	p_gather##suffix(&(dummy[0]), R_##suffix); \
	ENDT("void p_gather" #suffix "(p_" #suffix " *, p_" #suffix ")")
P_ALPHA
#undef	P_DECLARE

	BEGINT
	p_8u dummy[TABSIZE];
	p_gatherz(&(dummy[0]), &(tab_8u[0]), (TABSIZE / NPROC));
	ENDT("void p_bcastPutz(ptr, ptr, int (1024 / NPROC))")

#define	P_DECLARE(suffix) \
	BEGINT \
	if (R_only1) { \
		p_scatPut##suffix(&(tab_##suffix[0])); \
	} else { \
		p_scatGet##suffix(); \
	} \
	ENDT("p_" #suffix " p_scatPut" #suffix "(p_" #suffix " *) / p_" #suffix " p_scatGet" #suffix "()")
P_ALPHA
#undef	P_DECLARE

	BEGINT
	p_8u dummy[TABSIZE];
	if (R_only1) {
		p_scatPutz(&(dummy[0]), &(tab_8u[0]), (TABSIZE / NPROC));
	} else {
		p_scatGetz(&(dummy[0]), (TABSIZE / NPROC));
	}
	ENDT("void p_scatPutz(ptr, ptr, int (1024 / NPROC)) / void p_scatGetz(ptr, int (1024 / NPROC))")

#define	P_DECLARE(suffix) \
	BEGINT \
	p_##suffix dummy[NPROC]; \
	p_all2all##suffix(&(dummy[0]), &(tab_##suffix[0])); \
	ENDT("p_all2all" #suffix "(p_" #suffix " *, p_" #suffix " *)")
P_ALPHA
#undef	P_DECLARE

	BEGINT
	p_8u dummy[TABSIZE];
	p_all2allz(&(dummy[0]), &(tab_8u[0]), (TABSIZE / NPROC));
	ENDT("void p_all2allz(ptr, ptr, int (1024 / NPROC))")

	putchar('\n');
}


void
reductions(void)
{
#define	P_DECLARE(suffix) \
	BEGINT \
	p_reduceAnd##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_reduceAnd" #suffix "(p_" #suffix ")")
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_reduceOr##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_reduceOr" #suffix "(p_" #suffix ")")
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_reduceXor##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_reduceXor" #suffix "(p_" #suffix ")")
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_reduceMin##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_reduceMin" #suffix "(p_" #suffix ")")
P_ALPHA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_reduceMax##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_reduceMax" #suffix "(p_" #suffix ")")
P_ALPHA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_reduceAdd##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_reduceAdd" #suffix "(p_" #suffix ")")
P_GAMMA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_reduceMul##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_reduceMul" #suffix "(p_" #suffix ")")
P_GAMMA
#undef	P_DECLARE

	putchar('\n');
}


void
scans(void)
{
#define	P_DECLARE(suffix) \
	BEGINT \
	p_scanAnd##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_scanAnd" #suffix "(p_" #suffix ")")
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_scanOr##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_scanOr" #suffix "(p_" #suffix ")")
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_scanXor##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_scanXor" #suffix "(p_" #suffix ")")
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_scanMin##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_scanMin" #suffix "(p_" #suffix ")")
P_ALPHA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_scanMax##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_scanMax" #suffix "(p_" #suffix ")")
P_ALPHA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_scanAdd##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_scanAdd" #suffix "(p_" #suffix ")")
P_GAMMA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_scanMul##suffix(R_##suffix); \
	ENDT("p_" #suffix " p_scanMul" #suffix "(p_" #suffix ")")
P_GAMMA
#undef	P_DECLARE

	putchar('\n');
}


void
voting(void)
{
	BEGINT
	p_first(R_1u);
	ENDT("int p_first(p_1u)");

	BEGINT
	p_count(R_1u);
	ENDT("int p_count(p_1u)");

	BEGINT
	p_quantify(R_1u);
	ENDT("int p_quantify(p_1u)");

	BEGINT
	p_voteCount(R_PEno);
	ENDT("int p_voteCount(int)");

	BEGINT
	p_vote(R_PEno);
	ENDT("p_bar p_vote(int)");

#define	P_DECLARE(suffix) \
	BEGINT \
	p_matchCount##suffix(R_##suffix); \
	ENDT("int p_matchCount" #suffix "(p_" #suffix ")")
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_match##suffix(R_##suffix); \
	ENDT("p_bar p_match" #suffix "(p_" #suffix ")")
P_BETA
#undef	P_DECLARE

#define	P_DECLARE(suffix) \
	BEGINT \
	p_rank##suffix(R_##suffix); \
	ENDT("int p_rank" #suffix "(p_" #suffix ")")
P_ALPHA
#undef	P_DECLARE

	putchar('\n');
}


void
scontrol(void)
{
	BEGINT
	s_ignore(R_bar);
	ENDT("void s_ignore(p_bar)")

	BEGINT
	s_install(S_SIGANY, s_ignore);
	ENDT("int s_install(int, void(*)(p_bar))")

	BEGINTS
	s_exit(); s_init();
	ENDTS("void s_exit(void); int s_init(void)")

	putchar('\n');
}


void
sprimitives(void)
{
	register int t = s_install(S_SIGANY, s_ignore);

	BEGINTS
	if (R_1u) s_signal(t);
	ENDTS("void s_signal(int)")

	BEGINTS
	s__poll();
	ENDTS("void s__poll(void)")

	BEGINTS
	s_poll();
	ENDTS("void s_poll(void)")

	BEGINTS
	s__wait();
	ENDTS("void s__wait(void)")

	BEGINTS
	s_wait();
	ENDTS("void s_wait(void)")

	putchar('\n');
}


int
main(register int argc,
register char **argv,
register char **arge)
{
	register p_64u seed = 0;
	register p_32 count = 1;

	if ((argc < 2) || (argc > 3)) {
		fprintf(stderr,
			"Usage: %s seed [repeat count]\n",
			argv[0]);
		exit(1);
	}

	/* check-in with AFAPI */
	if (p_init()) exit(1);

	if (IPROC == CPROC) {
		seed = atoi(argv[1]);
		p_bcastPut64u(seed);
		if (argc == 3) count = atoi(argv[2]);
		p_bcastPut32(count);
	} else {
		seed = p_bcastGet64u();
		count = p_bcastGet32();
	}

	/* Announce what we are testing... */
	printf("%s AFAPI\t%s\n", AFAPI, AFAPI_VERSION);
	printf("Random seed\t%u\n", seed);
	printf("NPROC\t%d\n", NPROC);

	/* init random generators */
	p_randInit(seed);

	/* Initialize and calibrate everything */
	setup_tabs();
	calibrate();
	scalibrate();
	putchar('\n');

	/* run the tests */
	while (count > 0) {
		/* Test all the p_ stuff */
		control();
		primitives();
		communications();
		reductions();
		scans();
		voting();

		/* Test the basic s_ stuff */
		scontrol();
		sprimitives();

		--count;
	}

	/* Exit cleanly */
	p_exit();
	exit(0);
}
