/*	rand.c

	AFAPI-based pseudo-RANDom generators.

	Two different types of random generators are provided.  All the
	random generators are initialized by calling p_randInit(x)
	where x is the same p_64u seed value for all processors.  The
	p_randUniq*() generators create a unique random sequence for
	each processor.  In contrast, the p_randSame*() generators
	create the exact same sequence on all processors (provided
	that the routines are called in the same order on every
	processor) -- this is particularly useful because these
	routines DO NOT PERFORM ANY COMMUNICATIONS, and thus have far
	less overhead than broadcasting from a single generator.

	Although the concept of a random pointer (p_ptr) value is a bit
	scary, the functions are defined for all the types in the
	P_ALPHA type set.  The value returned is between 0 and 1.0 for
	floating point functions; all bits are used for the functions
	that return integer values, so signed integer values can be
	negative.  The distribution is, at least in theory, flat.

	This code is based on the approach outlined in Chapter 10 of
	"Designing and Building Parallel Programs," by Ian Foster.
	In section 10.1 he suggests a formula equivalent to:

	rand = (a * rand) % m

	and further states that the repeat period is m/4 for the case
	where m is a power of two and a is 8 * k + 5 where k is an
	integer.  The p_randSame*() functions use precisely this
	method.  In section 10.3.2, he describes "the Leapfrog method"
	for converting the above serial formula into a set of parallel
	generators interleaving in the original pseudo-random
	sequence.  This technique is used to construct the
	p_randUniq*() functions.

	However, taken by themselves, the low bits of these
	pseudo-random numbers are not really very random.  Thus, in
	constructing the random numbers, a larger random value is
	generated and then that value is shifted right somewhat to
	increase randomness of the low bits of the results returned.

	September 1997 by H. Dietz
*/


#include <AFAPI/afapi.h>

#define	P_RANDMULT32	((p_32u) ((8 * 16807) + 5))
#define	P_RANDMULT64	((p_64u) ((8 * ((p_64u) 33952834046453LL)) + 5))

static p_64u p_randUSeed64;
static p_64u p_randUMult64;
static p_32u p_randUSeed32;
static p_32u p_randUMult32;

static p_64u p_randSeed64;
static p_32u p_randSeed32;

void
p_randInit(register p_64u seed)
{
	register int i;

	/* Set p_randSame seeds according to seed... */
	p_randSeed32 = ~(p_randSeed64 = seed);

	/* Set p_randUniq seeds according to IPROC... */
	p_randUSeed32 = ~(p_randUSeed64 = (seed * (IPROC + 1)));

	/* Now skew the seed so that PE k has the kth value in an
	   imaginary sequential random sequence of 64-bit values with
	   a 62-bit repetition length, and 32-bit values with a
	   sequence length of 30 bits...
	*/
	for (i=0; i<IPROC; ++i) {
		p_randUSeed64 *= P_RANDMULT64;
		p_randUSeed32 *= P_RANDMULT32;
	}

	/* The parallel random generator should have each PE sample
	   the NPROCth next value in the sequential random sequence.
	   This is done by using P_RANDMULT to the NPROC power as the
	   multiplier...  so here we compute this value.
	*/
	p_randUMult64 = P_RANDMULT64;
	p_randUMult32 = P_RANDMULT32;
	for (i=1; i<NPROC; ++i) {
		p_randUMult64 *= P_RANDMULT64;
		p_randUMult32 *= P_RANDMULT32;
	}
}


#define	P_DECLARE(suffix) \
p_##suffix \
p_randUniq##suffix(void) \
{ \
	return((p_##suffix) \
	       (((p_randUSeed32 *= p_randUMult32) >> 8) & \
	        ((((p_32u) 1) << p_##suffix##_bits) - 1))); \
}

P_DECLARE(1u)
P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)

#undef	P_DECLARE

#define	P_DECLARE(suffix) \
p_##suffix \
p_randUniq##suffix(void) \
{ \
	return((p_##suffix) \
	       (((p_randUSeed64 *= p_randUMult64) >> 16) & \
	        ((((p_64u) 1) << p_##suffix##_bits) - 1))); \
}

P_DECLARE(bar)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(ptr)

#undef	P_DECLARE

p_64u
p_randUniq64u(void)
{
	return((p_randUSeed64 *= p_randUMult64) ^
	       (((p_64u) (p_randUSeed32 *= p_randUMult32)) << 32));
}

p_64
p_randUniq64(void)
{
	return((p_64u) p_randUniq64u());
}

p_64f
p_randUniq64f(void)
{
	return(((p_64f) p_randUniq64u()) /
	       ((p_64f) ((p_64u) (((p_64u) 0) - 1))));
}

p_32f
p_randUniq32f(void)
{
	return(((p_32f) p_randUniq32u()) /
	       ((p_32f) ((p_32u) (((p_32u) 0) - 1))));
}


#define	P_DECLARE(suffix) \
p_##suffix \
p_randSame##suffix(void) \
{ \
	return((p_##suffix) \
	       (((p_randSeed32 *= P_RANDMULT32) >> 8) & \
	        ((((p_32u) 1) << p_##suffix##_bits) - 1))); \
}

P_DECLARE(1u)
P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)

#undef	P_DECLARE

#define	P_DECLARE(suffix) \
p_##suffix \
p_randSame##suffix(void) \
{ \
	return((p_##suffix) \
	       (((p_randSeed64 *= P_RANDMULT64) >> 16) & \
	        ((((p_64u) 1) << p_##suffix##_bits) - 1))); \
}

P_DECLARE(bar)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(ptr)

#undef	P_DECLARE

p_64u
p_randSame64u(void)
{
	return((p_randSeed64 *= P_RANDMULT64) ^
	       (((p_64u) (p_randSeed32 *= P_RANDMULT32)) << 32));
}

p_64
p_randSame64(void)
{
	return((p_64u) p_randSame64u());
}

p_64f
p_randSame64f(void)
{
	return(((p_64f) p_randSame64u()) /
	       ((p_64f) ((p_64u) (((p_64u) 0) - 1))));
}

p_32f
p_randSame32f(void)
{
	return(((p_32f) p_randSame32u()) /
	       ((p_32f) ((p_32u) (((p_32u) 0) - 1))));
}
