/*	intern.h

	AFAPI internals for SHMAPERS.

	September 1998 by H. Dietz and T.I.Mattox
*/
#if	NPROC == 1
#include "../ONE/one.h"
#endif
#if	NPROC > 1

#ifndef _P_INTERN_H
#define _P_INTERN_H

#include <AFAPI/afapi.h>

/*	SHM data structure stuff...

	Notice that this attempts to add padding to keep things
	aligned with cache lines so that info for each PE can be
	updated independently without "false sharing" overhead.  This
	also turns the non-atomic spin locks into spins on local
	cache, whereas atomic spin locks would suck bus cycles,
	slowing down other processors and I/O devices.

	Also note that the pointers to shared memory are not volatile;
	only the stuff in shared memory is volatile.
*/
#define	P_DECLARE(suffix) p_##suffix u_##suffix;
union p_shm_union {
	P_ALPHA	/* An instance of each AFAPI type */
};
#undef	P_DECLARE

typedef struct {
	union p_shm_union datum[2];	/* Double buffered */
	int	barno;			/* Barrier number for this PE */
	short	pid;			/* UNIX PID of this PE */

	/* Add padding to make a multiple of cache line size */
	char	pad[P_CACHELINE -
		    ((sizeof(int) +
		      sizeof(short) +
		      sizeof(union p_shm_union) +
		      sizeof(union p_shm_union)) % P_CACHELINE)];
} p_shm_type;

extern volatile p_shm_type *p_shm_all;	/* Pointer to all PE info */
extern volatile p_shm_type *p_shm_me;	/* Pointer to my PE info */

#define	P_BLOCKSIZE (P_CACHELINE * 32)	/* Broadcast block size */

extern int p_shm_id;			/* System V shared segment ID */

#ifndef	P_SHM_IDLE
#if	NPROC<=SMPPROC
#define	P_SHM_IDLE	{}		/* NPROC real PEs */
#else
#include <unistd.h>			/* Where usleep() is declared */
#define	P_SHM_IDLE	usleep(1)	/* NPROC virtual PEs */
#endif
#endif


/*
	AFAPI SHMAPERS inline primitive functions....

	Unlike other versions of AFAPI, this library is intended to
	obtain maximum performance from stock hardware.  In particular,
	this is optimized for SMP machines that support System V
	shared memory (which is generally grouped with the System V
	IPC support...  make sure you have it enabled in your kernel).
	If you are not sure, try typing "ipcs"; if you get a status
	report on shared memory, semaphores, and message queues, you
	should be fine.

	Note that, although this library will run on single-processor
	machines, it will spend a lot of time spinning that way.

	July 1996 by Hank Dietz
*/



/*	p__wait()...


	There are lots of ways to barrier synchronize using shared
	memory, including the use of various atomic operations.
	However, atomic spin locks can truly swamp the bus.

	In contrast, this code uses no atomic locks.  Basically, the
	idea is that reads from a shared object can be cached, hence
	they will not cause bus traffic until another processor
	invalidates.  The use of a counter for each PE simplifies the
	code and also allows a two-phase rather than a four-phase
	barrier.  There is also the minor optimization that if you
	find any PE is waiting for the next barrier, then he must have
	already found everyone at this barrier, so I'm done
	immediately.

	The final optimization is the insertion of a P_SHM_IDLE as a
	suggestion to the UNIX scheduler that this job may be
	spin-locked for some time (i.e., schedule another job).  In
	types.h, P_SHM_IDLE should be set to usleep(1) for heavily
	loaded systems or systems with slow coherence mechanisms; for
	fast, lightly-loaded, systems, define it to be nothing.
*/

extern inline
void
p__wait(void)
{
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = ++(p_shm_me->barno);
	register int barnonext = (barno + 1);

	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) return;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) return;
			} while (hisbarno != barno);
		}
	} while (++p < q);
}

/*	p__waitbar(), p__any(), and p__all()...
*/

extern inline
p_bar
p__waitbar(register p_1u f)
{
	/* Collect the array of flags and compress it to bits... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);
	register p_bar result = 0;

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Convert value array into packed bits... */
	p = p_shm_all;
	do {
		--q;
		result = ((result << 1) | q->datum[ which ].u_1u);
	} while (q > p);

	return(result);
}


extern inline
p_1u
p__any(register p_1u f)
{
	/* Collect the array of flags and look for a 1... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Scan for a 1... */
	p = p_shm_all;
	do {
		if (p->datum[ which ].u_1u) return(1);
	} while (++p < q);

	return(0);
}


extern inline
p_1u
p__all(register p_1u f)
{
	/* Collect the array of flags and look for a 0... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Scan for a 0... */
	p = p_shm_all;
	do {
		if (0 == p->datum[ which ].u_1u) return(0);
	} while (++p < q);

	return(1);
}

/*	Inline signal primitives...

	Signals are really very easy because we use the barrier
	number.  If somebody is at the next count, we know that
	they are sending a signal....
*/

extern inline void
s__signal(register int s)
{
	/* This is not worth inlining... */
	s_signal(s);
}


extern inline p_1u
s__pending(void)
{
	return(s_pending());
}


extern inline void
s__poll(void)
{
	if (s__pending()) s_signal(S_SIGANY);
}


extern inline void
s__wait(void)
{
	/* Just exit and re-enter... that flushes things */
	s_exit();
	s_init();
}


#endif
#endif /* if NPROC > 1 */
