/*	vwmain.c

	Video Wall MAIN function and display setup,
	including stuff dealing with wall config.

	This code reads a configuration file to define
	a set of video walls.  Syntax of the file is:

	file:  (WALLNAME ':' dim '{' (panel)* '}')*
	panel: PENUMBER ':' dim
	dim:   XMIN YMIN XMAX YMAX

	March 1999, H. Dietz
*/


#include <AFAPI/vwsup.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>


static vwwall allwalls[VWMAXWALLS];
static int nextwall = 0;
static int fd = -1;
static int tok = ' ';


static void
advance(void)
{
	if (fd != -1) {
		char c;
		if (read(fd, &c, 1) != 1) {
			close(fd);
			fd = -1;
			tok = -1;
		} else {
			tok = (c & 0x7f);
		}
	}
}

static void
getspace(void)
{
	for (;;) {
		while ((tok > 0) && ((tok <= ' ') ||
				     (tok == ';') ||
				     (tok == ','))) {
			advance();
		}

		/* also ignore comments between # and EOL */
		if (tok != '#') return;
		while ((tok > 0) && (tok != '\n')) {
			advance();
		}
	}
}

static double
getfloat(void)
{
	register double t = 0.0;

	getspace();
	while (isdigit(tok)) {
		t *= 10.0;
		t += (tok - '0');
		advance();
	}
	if (tok == '.') {
		register double scale = 1.0;

		advance();
		while (isdigit(tok)) {
			scale /= 10.0;
			t += (scale * (tok - '0'));
			advance();
		}
	}

	return(t);
}

static void
getdims(register vwfbox *p)
{
	if (p) {
		p->min.x = getfloat();
		p->min.y = getfloat();
		p->max.x = getfloat();
		p->max.y = getfloat();
	} else {
		getfloat();
		getfloat();
		getfloat();
		getfloat();
	}
}

/* Parse as many wall specs as we recognize in file 'filename' */
void
vwdefWalls(register char *filename)
{
	/* Open the file */
	if (filename == 0) return;
	if ((fd = open(filename, O_RDONLY)) < 0) return;

	getspace();
	while (isalpha(tok) && (nextwall < VWMAXWALLS)) {
		register char *p = ((char *) malloc(32));
		register int i;
		register int out_of_range = 0;

		allwalls[nextwall].name = p;
		do {
			if ((p - allwalls[nextwall].name) < 30) {
				*(p++) = tok;
			}
			advance();
		} while (isalpha(tok) || isdigit(tok));
		*p = 0;
		getspace();
		if (tok == ':') advance();

		getdims(&(allwalls[nextwall].all));
		for (i=0; i<NPROC; ++i) {
			allwalls[nextwall].one[i].min.x = 0;
			allwalls[nextwall].one[i].min.y = 0;
			allwalls[nextwall].one[i].max.x = 0;
			allwalls[nextwall].one[i].max.y = 0;
		}
		getspace();
		if (tok != '{') {
			fprintf(stderr,
"Warning: Malformed definition for wall '%s' in file '%s'.\n",
				allwalls[nextwall].name, filename);
			break;
		}
		advance();
		getspace();
		while (isdigit(tok)) {
			/* Parse PE_number: */
			i = 0;
			while (isdigit(tok)) {
				i *= 10;
				i += (tok - '0');
				advance();
			}
			getspace();
			if (tok == ':') advance();

		/* Now parse clip area, ignorning PEs out of range */
			if ((i >= 0) && (i < NPROC)) {
				getdims(&(allwalls[nextwall].one[i]));
			} else {
				++out_of_range;
				getdims((void *) 0);
			}
			getspace();
		}
		if (out_of_range) {
			fprintf(stderr,
"Warning: %d nonexistent PE(s) were listed for wall '%s'\n"
"         in file '%s'.\n",
	out_of_range, allwalls[nextwall].name, filename);
		}
		if (tok != '}') {
			fprintf(stderr,
"Warning: Malformed definition for wall '%s' in file '%s'.\n",
				allwalls[nextwall].name, filename);
			break;
		}
		advance();

		++nextwall;
		getspace();
	}
	if ((nextwall == VWMAXWALLS) && isalpha(tok)) {
		fprintf(stderr,
"Warning: At most %d named walls can be defined.  That limit was reached when\n"
"         parsing file '%s'.\n",
			VWMAXWALLS, filename);
	}

	/* Clean up if necessary */
	tok = ' ';
	if (fd != -1) {
		close(fd);
		fd = -1;
	}
}


vwwall *
vwgetWall(register char *name)
{
	/* Find most recent definition of named wall */
	register int i;

	for (i=(nextwall - 1); i>=0; --i) {
		if (strcmp(allwalls[i].name, name) == 0) {
			return(&(allwalls[i]));
		}
	}
	return(0);
}


void
vwdefSpace(register vwwall *wall)
{
	register int i;

	if (wall == 0) {
		for (i=0; i<NPROC; ++i) {
			vwspace[i] = 0;
		}
	} else {
		for (i=0; i<NPROC; ++i) {
			if (wall->one[i].min.x != wall->one[i].max.x) {
				vwspace[i] = wall;
			}
		}
	}
}

				

void
vwinit(void)
{
	register int i;

	/* Check in */
	if (p_init()) exit(1);

	/* Pre-install default display */
	allwalls[0].name = "default";
	allwalls[0].all.min.x = 0.0;
	allwalls[0].all.min.y = 0.0;
	allwalls[0].all.max.x = 12.0;
	allwalls[0].all.max.y = 9.0;
	for (i=0; i<NPROC; ++i) {
		allwalls[0].one[i] = allwalls[0].all;
	}
	nextwall = 1;
	vwdefWalls(getenv("VWSPACE"));
	vwdefSpace(0);
	vwdefSpace(vwgetWall("default"));
}

void
vwexit(void)
{
	if (vwspace[IPROC]) {
		vwend();
	}
	vwendMouse();
	vwendKeybrd();

	/* Check out */
	p_exit();
}

inline void
vpclear(register vwbuf *panel)
{
	/* Clear this PE's panel buffer */
	memset(panel->data.data,
	       0, /* assumes that 0 is black */
	       panel->bytes);

	/* And it is all "dirty" */
	panel->dirty.min.x = 0;
	panel->dirty.min.y = 0;
	panel->dirty.max.x = panel->dim.x - 1;
	panel->dirty.max.y = panel->dim.y - 1;
}

void
vwclear(register vwwall *wall)
{
	/* vwclear(0) clears my panel no matter what;
	   vwclear(wall) clears my panel only if it's in wall
	*/
	if (((!wall) && vwspace[IPROC]) ||
	    (wall == vwspace[IPROC])) {
		vpclear(&vwpanel);
	}
}
