/*	vwppm.c

	Video Wall PPM support.

	Currently, this supports only 24-bit color raw PPM
	files, the P6 format.  This is because the other PPM
	formats are not able to be accessed directly as 24-bit
	image data, so we would need to make a copy of the data
	rather than simply mapping it into memory.

	February 1999 by H. Dietz
*/


#include <AFAPI/vwsup.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <malloc.h>


#define	MAXOPEN	32

static struct ppminfo {
	vwbuf	image;		/* image info */

	/* Fields below are internal use only...
	*/
	int	maxval;		/* maximum value */
	int	refcnt;		/* how many refs? */
	int	fd;		/* file descriptor */
	void	*map;		/* map address */
	int	fsize;		/* mapped file size */
	char	*fname;		/* file name */
} vwppmInfo[MAXOPEN];

static int vwppminit = 0;

extern int vpwriteFile(register char *filename, vwbuf image);

vwbuf
vpmapFile(register char *filename)
{
	/* The following is a fast and sloppy way to
	   read a color raw PPM (P6) image file
	*/
	register int fd;
	register int fsize;
	register unsigned char *map;
	register unsigned char *p;
	register struct ppminfo *info;
	vwbuf empty;

	/* Init? */
	if (!vwppminit) {
		for (info=&(vwppmInfo[0]); info<&(vwppmInfo[MAXOPEN]); ++info) {
			info->image.depth = 0;
			info->refcnt = 0;
			info->fname = 0;
		}
		vwppminit = 1;
	}

	/* Is this file mapped already? */
	for (info=&(vwppmInfo[0]); info<&(vwppmInfo[MAXOPEN]); ++info) {
		if ((info->fname != 0) &&
		    !strcmp(info->fname, filename)) {
			++(info->refcnt);
			return(info->image);
		}
	}

	/* Find an empty struct ppminfo */
	empty.dim.x = 0;
	empty.dim.y = 0;
	empty.dirty.min.x = -1;
	empty.dirty.min.y = -1;
	empty.dirty.max.x = -1;
	empty.dirty.max.y = -1;
	empty.depth = 0;
	empty.data.data = 0;
	for (info=&(vwppmInfo[0]); info<&(vwppmInfo[MAXOPEN]); ++info) {
		if (info->fname == 0) {
			goto gotslot;
		}
	}
	return(empty);

gotslot:
	info->image = empty;

	/* First, open the file... */
	if ((fd = open(filename, O_RDONLY)) < 0) {
		return(empty);
	}

	/* Read size and map the whole file... */
	fsize = lseek(fd, ((off_t) 0), SEEK_END);
	map = ((unsigned char *)
	       mmap(0,		/* Put it anywhere */
		    fsize,	/* Map the whole file */
		    PROT_READ,	/* Read only */
		    MAP_SHARED,	/* Not just for me */
		    fd,		/* The file */
		    0));	/* Right from the start */
	if (((int) map) == -1) {
		close(fd);
		return(empty);
	}

	/* File should now be mapped; read magic value */
	p = map;
	if ((*(p++) != 'P') || (*(p++) != '6')) {
		/* It isn't a P6 PPM file...
		   try to convert it and then map that.
		*/
		char name[32];
		char buf[512];
		close(fd);
		munmap(map, fsize);
		sprintf(&(name[0]), "vw%02x.ppm", (info - &(vwppmInfo[0])));
		sprintf(&(buf[0]),
			VWCONVERT " %s %s",
			filename,
			&(name[0]));
		if (system(&(buf[0])) != 0) {
			/* Something failed... */
			unlink(&(name[0]));
			return(empty);
		}

		/* Open the temporary file... */
		if ((fd = open(&(name[0]), O_RDONLY)) < 0) {
			unlink(&(name[0]));
			return(empty);
		}

		/* Unlink the name so it disappears when closed */
		unlink(&(name[0]));

		/* Read size and map the whole file... */
		fsize = lseek(fd, ((off_t) 0), SEEK_END);
		map = ((unsigned char *)
		       mmap(0,		/* Put it anywhere */
			    fsize,	/* Map the whole file */
			    PROT_READ,	/* Read only */
			    MAP_SHARED,	/* Not just for me */
			    fd,		/* The file */
			    0));	/* Right from the start */
		if (((int) map) == -1) {
			close(fd);
			return(empty);
		}

		/* File should now be mapped; read magic value */
		p = map;
		if ((*(p++) != 'P') || (*(p++) != '6')) {
			close(fd);
			return(empty);
		}
	}

#define	Eat_Space \
	while ((*p == ' ') || \
	       (*p == '\t') || \
	       (*p == '\n') || \
	       (*p == '\r') || \
	       (*p == '#')) { \
		if (*p == '#') while (*(++p) != '\n') ; \
		++p; \
	}

	Eat_Space;		/* Eat white space and comments */

#define	Get_Number(n) \
	{ \
		register int charval = *p; \
 \
		if ((charval < '0') || (charval > '9')) goto ppm_exit; \
 \
		n = (charval - '0'); \
		charval = *(++p); \
		while ((charval >= '0') && (charval <= '9')) { \
			n *= 10; \
			n += (charval - '0'); \
			charval = *(++p); \
		} \
	}

	Get_Number(info->image.dim.x);	/* Get image width */

	Eat_Space;			/* Eat white space and comments */
	Get_Number(info->image.dim.y);	/* Get image width */

	Eat_Space;			/* Eat white space and comments */
	Get_Number(info->maxval);	/* Get image max value */

	/* Should be 8-bit binary after one whitespace char... */
	if (info->maxval > 255) {
ppm_exit:
		close(fd);
		munmap(map, fsize);
		return(empty);
	}
	if ((*p != ' ') &&
	    (*p != '\t') &&
	    (*p != '\n') &&
	    (*p != '\r')) goto ppm_exit;

	/* Here we are... next byte begins the 24-bit data */
	info->image.dirty.min.x = 0;
	info->image.dirty.min.y = 0;
	info->image.dirty.max.x = info->image.dim.x - 1;
	info->image.dirty.max.y = info->image.dim.y - 1;
	info->image.depth = 24;
	info->image.bytes = (3 * info->image.dim.x * info->image.dim.y);
	info->image.data.data24 = (p + 1);
	info->fd = fd;
	info->map = map;
	info->fsize = fsize;
	info->fname = malloc(strlen(filename));
	strcpy(info->fname, filename);
	info->refcnt = 1;
	return(info->image);
}

#undef	Eat_Space
#undef	Get_Number


vwbuf
vwmapFile(register char *filename)
{
	vwbuf image = vpmapFile(filename);

	/* Does any PE not have it local? */
	if (p_any(image.bytes == 0)) {
		/* Whose copy to broadcast?
		   Prefer CPROC, else lowest PE number.
		*/
		register int src = CPROC;
		if (!p_any((IPROC == CPROC) && image.bytes)) {
			src = p_first(image.bytes != 0);
		}

		/* Broadcast the file */
		if (IPROC == src) {
			/* Sender */
			p_bcastPut32(image.dim.x);
			p_bcastPut32(image.dim.y);
			p_bcastPut32(image.depth);
			p_bcastPut32(image.bytes);
			p_bcastPutz(image.data.data, image.bytes);
		} else if (image.bytes == 0) {
			/* PE that needs the file...
			   Get a memory copy of it,
			   write it to disk,
			   free the copy,
			   map the file,
			   and unlink the file name.
			   This makes the image swap to the file,
			   thus not wasting valuable swap space,
			   yet the file will go away with the
			   termination of this program or mapping.
			*/
			image.dim.x = p_bcastGet32();
			image.dim.y = p_bcastGet32();
			image.depth = p_bcastGet32();
			image.bytes = p_bcastGet32();
			image.data.data = malloc(image.bytes);
			p_bcastGetz(image.data.data, image.bytes);
			vpwriteFile(filename, image);
			free(image.data.data);
			image = vpmapFile(filename);
			unlink(filename);
		} else {
			/* PE not sending, but that has file */
			p_bcastGet32();
			p_bcastGet32();
			p_bcastGet32();
			p_bcastGetz(0, p_bcastGet32());
		}
	}

	return(image);
}


void
vpunmapFile(register char *filename)
{
	register struct ppminfo *info;

	/* Is this file mapped already? */
	for (info=&(vwppmInfo[0]); info<&(vwppmInfo[MAXOPEN]); ++info) {
		if ((info->fname != 0) &&
		    !strcmp(info->fname, filename)) {
			if (--(info->refcnt) == 0) {
				free(info->fname);
				info->fname = 0;
				close(info->fd);
				munmap(info->map, info->fsize);
				info->image.depth = 0;
			}
			return;
		}
	}
}


void
vwunmapFile(register char *filename)
{
	vpunmapFile(filename);
}


int
vpwriteFile(register char *filename,
vwbuf image)
{
	register int fd;
	register int len;
	char buf[512];

	/* Can only write 24-bit data */
	if (image.depth != 24) return(1);

	/* Open/create the file... */
	fd = open(filename, (O_WRONLY | O_CREAT), 0666);
	if (fd < 0) {
		return(2);
	}

	/* Then write the header...
	   the header size is a multiple of 16 for alignment
	*/
	sprintf(&(buf[0]),
		"P6\n#%21.21s\n%8d %8d\n255\n",
		filename,
		image.dim.x,
		image.dim.y);
	if (write(fd, &(buf[0]), 48) != 48) {
		close(fd);
		return(2);
	}

	/* Then write the data... */
	len = (image.dim.x * image.dim.y * 3);
	if (write(fd, image.data.data, len) != len) {
		close(fd);
		return(3);
	}

	/* All looks good */
	close(fd);
	return(0);
}
