/*	vwtime.c

	Video Wall TIME routines.

	There are two completely different versions; one uses
	standard unix time, the other uses the Intel performance
	register clock counter, which we assume is not
	manipulated by any user processes nor system tasks.  The
	unix clock version is used if VWINTELTICKS is not
	defined in vwtime.h; the helper vwtimehelp.c is used to
	determine the actual clock frequency by counting ticks
	over a relatively long period of time referenced by unix
	timing methods.

	February 1999, by H. Dietz
*/


#include <unistd.h>
#include <AFAPI/vwsup.h>
#include <AFAPI/vwtime.h>


/*	If times not provided by vwtimehelp, define approximate
	costs of performing usleep(1) and vwnow()
*/

#ifndef	VWUSLEEP1TIME
#define	VWUSLEEP1TIME	(150.0 / 1000000.0)
#endif	VWUSLEEP1TIME

#ifndef	VWNOWTIME
#define	VWNOWTIME	(15.0 / 1000000.0)
#endif	VWNOWTIME

/*	How closely should we be able to sync?
	If using AFAPI, no closer than 3 useconds.
*/
#if	NPROC > 1
#define	VWERRTIME	(((3 * VWNOWTIME) < 3) ? 3 : ((int) (3 * VWNOWTIME)))
#else
#define	VWERRTIME	((int) (3 * VWNOWTIME))
#endif

#ifndef	VWINTELTICKS

#include <time.h>
#include <sys/times.h>
#include <sys/time.h>

static	struct timeval	sysref;

vwtime
vwnow(void)
{
	static struct timeval sysnow;
	register vwtime mynow;

	gettimeofday(&sysnow, 0);
	mynow = ((1000000.0 * (sysnow.tv_sec - sysref.tv_sec)) +
		 (sysnow.tv_usec - sysref.tv_usec));
	return(mynow * (1.0 / 1000000.0));
}

void
vwreftime(void)
{
#if	NPROC > 1
	/* Calibrate global time within VWERRTIME us */
	register vwtime mynow;
	register vwtime minnow;

	do {
		p_wait();	/* get everybody here */
		usleep(1);	/* get us a clean slice */
		p_wait();	/* precise sync */
		gettimeofday(&sysref, 0);
		p_wait();	/* does nearly nothing */
		mynow = vwnow();
		minnow = p_reduceMin64f(mynow);
	} while (p_any(mynow > (minnow + VWERRTIME)));
#else
	gettimeofday(&sysref, 0);
#endif
}

#else	VWINTELTICKS

union tick_union {
	unsigned long long	val64;
	struct {
		unsigned int	low32;
		unsigned int	hi32;
	} lowhi;
};

static unsigned long long sysref = 0.0;

static unsigned long long
vwnowTick(void)
{
	register union tick_union result;
	register unsigned int msr = 0;

	__asm __volatile(
".byte 0xf; .byte 0x31	# Read the cycle counter
	movl %%edx, %0	# high order bits
	movl %%eax, %1	# low order bits"
: "=g" (result.lowhi.hi32), "=g" (result.lowhi.low32)
: "g" (msr)
: "eax", "edx"
			 );

	return(result.val64 - sysref);
}

vwtime
vwnow(void)
{
	return((vwtime) (vwnowTick() * (1.0 / VWINTELTICKS)));
}

void
vwreftime(void)
{
#if	NPROC > 1
	/* Calibrate global time within VWERRTIME us */
	register vwtime mynow;
	register vwtime minnow;

	do {
		p_wait();	/* get everybody here */
		sysref = 0LL;
		usleep(1);	/* get us a clean slice */
		p_wait();	/* precise sync */
		sysref = vwnowTick();
		p_wait();	/* does nearly nothing */
		mynow = vwnow();
		minnow = p_reduceMin64f(mynow);
	} while (p_any(mynow > (minnow + VWERRTIME)));
#else
	sysref = 0LL;
	sysref = vwnowTick();
#endif
}

#endif	VWINTELTICKS


int
vwwait(register vwtime deadline)
{
	if (deadline != 0.0) {
		register vwtime now;

		deadline -= VWERRTIME;
		now = vwnow();
		if ((now + ((int) (2 * VWUSLEEP1TIME + VWNOWTIME))) <
		    deadline) {
			usleep(deadline - (now - ((int) (2 * VWUSLEEP1TIME))));
			now = vwnow();
		}
		while ((now + ((int) (2 * VWNOWTIME))) < deadline) {
			now = vwnow();
		}

		if (now > (deadline + VWERRTIME)) {
			return(VWLATE);
		}
	}

	return(VWONTIME);
}
