/*	vwvga.c

	Video Wall VGA library support.

	March 1999, Hank Dietz
*/


#include <AFAPI/vwsup.h>
#include <vga.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <malloc.h>

#ifndef	VWVGA_BLIT

/*	Use deliberately random order to paint pixels;
	this minimizes the effects of the race between
	display retrace and repaint.  Blits are faster,
	but have nasty artifacts.
*/
static	int vworder[VWWIDE * VWTALL];

#endif	VWVGA_BLIT

void
vwbegin(void)
{
	register int i;

	/* Hack to get around VGA library checks that we own
	   the VGA display...  it would be better to simply
	   remove the VGA library check code, but....
	*/
	{
		register char *the_tty_name = ttyname(0);

		if ((the_tty_name == NULL) ||
		    (strcmp("/dev/tty1", the_tty_name) != 0)) {
			close(0); close(1); close(2);
			open("/dev/tty1", O_RDWR);
			open("/dev/tty1", O_RDWR);
			open("/dev/tty1", O_RDWR);
			setsid();
			ioctl(0, TIOCSCTTY, 0);
		}
	}

	vga_init();

	/* Select highest-res mode available */
	if ((VWWIDE >= 1024) &&
	    (VWTALL >= 768) && 
	    vga_hasmode(G1024x768x256)) {
		vga_setmode(G1024x768x256);
		vwpanel.dim.x = 1024;
		vwpanel.dim.y = 768;
	} else if ((VWWIDE >= 800) &&
		   (VWTALL >= 600) &&
		   vga_hasmode(G800x600x256)) {
		vga_setmode(G800x600x256);
		vwpanel.dim.x = 800;
		vwpanel.dim.y = 600;
	} else if ((VWWIDE >= 640) &&
		   (VWTALL >= 480) &&
		   vga_hasmode(G640x480x256)) {
		vga_setmode(G640x480x256);
		vwpanel.dim.x = 640;
		vwpanel.dim.y = 480;
	} else if ((VWWIDE >= 320) &&
		   (VWTALL >= 200) &&
		   vga_hasmode(G320x200x256)) {
		vga_setmode(G320x200x256);
		vwpanel.dim.x = 320;
		vwpanel.dim.y = 200;
	} else {
		fprintf(stderr,
			"No appropriate 8 bit VGA mode available\n");
		exit(1);
	}
	vwpanel.depth = 8;
	vwpanel.bytes = (vwpanel.dim.x * vwpanel.dim.y);
	vwpanel.data.data = malloc(vwpanel.bytes);
	vpclear(&vwpanel);

	vga_disabledriverreport();
	vga_screenon();
	vga_setpage(0);
	vga_setdisplaystart(0);

	/* Set-up our fixed 256-color map...  we treat 8 bit
	   color as a "True Color" format, and use dithering to
	   help compensate for the limited color matching
	   abilities.  One could dynamically select color maps,
	   but that can yield very strange behaviour when
	   attempting to display portions of an image across
	   multiple displays.
	*/
	for (i=0; i<256; ++i) {
		vga_setpalette(i,
			       (((i) & 0007) << 5) >> 2,
			       (((i) & 0070) << 2) >> 2,
			       ((i) & 0300) >> 2);
	}

#ifndef	VWVGA_BLIT

	/* Construct pixel draw ordering array */
	for (i=(vwpanel.bytes - 1); i>=0; --i) {
		vworder[i] = i;
	}
	/* Exchange within video page */
	for (i=(vwpanel.bytes - 1); i>=0; --i) {
		register int j = ((i & ~0xffff) |
				  ((rand() >> 3) & 0xffff));
		register int t = vworder[i];
		while (j >= vwpanel.bytes) j = ((i & ~0xffff) |
				  ((rand() >> 3) & 0xffff));
		vworder[i] = vworder[j];
		vworder[j] = t;
	}
	/* Then jumble pseudo-scan lines across pages */
	for (i=0; i<vwpanel.dim.y; ++i) {
		register int j = ((rand() >> 5) % vwpanel.dim.y);
		register int iw = (i * vwpanel.dim.x);
		register int jw = (j * vwpanel.dim.x);
		register int k;

		for (k=0; k<vwpanel.dim.x; ++k) {
			register int t = vworder[iw + k];
			vworder[iw + k] = vworder[jw + k];
			vworder[jw + k] = t;
		}
	}

#endif	VWVGA_BLIT
}

inline void
vppaint(register vwbuf *panel)
{
	/* Write panel to display one page at a time */
	register int i = panel->bytes;
	register int j = 0;
	register pixel8 *p = panel->data.data8;

#ifdef	VWVGA_BLIT

	while (i > 0) {
		vga_setpage(j++);
		if (i > 0x10000) {
			memcpy(graph_mem, p, 0x10000);
			p += 0x10000;
			i -= 0x10000;
		} else {
			memcpy(graph_mem, p, i);
			i = 0;
		}
	}
	vga_setpage(0);

#else	VWVGA_BLIT

	/* Randomized copy update */
	{
		register int *ord = &(vworder[panel->bytes]);
		register int page = -1;

		while (ord > &(vworder[0])) {
			register int i = *(--ord);
			register int newpage;

			if (page != (newpage = (i & ~0xffff))) {
				page = newpage;
				vga_setpage(i >> 16);
			}
				graph_mem[i & 0xffff] = p[i];
		}
	}
	vga_setpage(0);

#endif	VWVGA_BLIT
}

void
vwpaint(register vwwall *wall)
{
	/* vwpaint(0) paints my panel no matter what;
	   vwpaint(wall) paints my panel only if it's in wall
	*/
	if (((!wall) && vwspace[IPROC]) ||
	    (wall == vwspace[IPROC])) {
		vppaint(&vwpanel);
	}
}

void
vwend(void)
{
	if (vwspace[IPROC] && vwpanel.data.data) {
		free(vwpanel.data.data);
	}
	vga_setmode(TEXT);
}
