/*	vwx.c

	Video Wall X11 driver.

	This version uses the X11R6 library with MIT-SHM
	extensions; if your X server doesn't have the shared
	memory extension support, this code will not work.  It
	seems that testing for support is funky, and the primary
	reason it wouldn't work would be that the X display
	being used isn't on the current machine -- which doesn't
	make sense for a video wall.

	Note: this code attempts to determine the X display
	dimensions, but XFree86 allows larger virtual displays
	to be hardware panned, and the virtual dimensions are
	what we get.  If the virtual display size is larger than
	physical, what you really want is the physical
	dimensions, but XFree86 doesn't seem to easily provide
	those.  The definitions of VWWIDE and VWTALL that are in
	AFAPI/config.h can be used to override the virtual
	sizing for everything but capturing the whole display,
	which is correctly done as the whole virtual display
	even if that is larger than VWWIDE by VWTALL.  The
	physical size used is the minimum of the virtual size or
	VWWIDE by VWTALL.  Keep in mind that moving the cursor
	away from the top left will still pan the virtual
	display area, so a better fix is to edit your XF86Config
	file so that virtual and physical sizes are the same.

	X11 is not exactly a simple programming model, and I
	have learned it mostly by observing other X11 codes.
	Thus, most of the following code is indirectly derived
	from a variety of other examples of X11 programming,
	especially xlockmore.  After all, the video wall panels
	are really nothing more than very strange screen savers.

	March 1998 by Hank Dietz
	Revised, March 1999 by Hank Dietz
*/


#include <AFAPI/vwsup.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xresource.h>
#include <X11/Xatom.h>
#include <X11/X.h>
#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>


/*	The following code deals with virtual root windows...
	a problem associated with some virtual window managers.
*/

inline extern Window
VirtualRootWindowOfScreen(Screen *screen)
{
	static Screen *save_screen = ((Screen *) 0);
	static Window root = ((Window) 0);

	if (screen != save_screen) {
		register unsigned int i;
		Display *dpy = DisplayOfScreen(screen);
		Atom __SWM_VROOT = None;
		Window rootReturn, parentReturn, *children;
		unsigned int numChildren;

		/* Get apparent root window */
		root = RootWindowOfScreen(screen);

		/* Try to find virtual root from children */
		__SWM_VROOT = XInternAtom(dpy, "__SWM_VROOT", False);
		if (XQueryTree(dpy,
			       root,
			       &rootReturn,
			       &parentReturn,
			       &children,
			       &numChildren)) {
			for (i=0; i<numChildren; ++i) {
				Atom actual_type;
				int actual_format;
				unsigned long nitems, bytesafter;
				Window *newRoot = (Window *) 0;

				if ((XGetWindowProperty(dpy,
					children[i],
					__SWM_VROOT,
					0,
					1,
					False,
					XA_WINDOW,
					&actual_type,
					&actual_format,
					&nitems,
					&bytesafter,
					((unsigned char **) &newRoot)) ==
					Success) && newRoot) {
					root = *newRoot;
					break;
				}
			}
			if (children) {
				XFree((char *) children);
			}
		}
		save_screen = screen;
	}
	return(root);
}



static Display *dsp = NULL;	/* server display connection */
static int screen;		/* current screen */
static GC my_gc;

static Window win;		/* window used to cover screen */

static Window root;		/* convenience pointer to the root window */
static Cursor mycursor;		/* blank cursor */
static Pixmap lockc;
static Pixmap lockm;		/* pixmaps for cursor and mask */
static char no_bits[] = {0};	/* dummy array for the blank cursor */
static int sstimeout;		/* screen saver parameters */
static int ssinterval;
static int ssblanking;
static int ssexposures;


XShmSegmentInfo SHMInfo;
static XImage *vwximage;

/*	real screen dim for mouse scaling */
static int realxdim, realydim;

void
vwbegin(void)
{
	XSetWindowAttributes xswa;
	XGCValues xgcv;
	XColor nullcolor;
	Screen *scr;
	Colormap cmap;
	char *displayname;

	displayname = ":0.0";
	if (!(dsp = XOpenDisplay(displayname))) {
		fprintf(stderr,"Unable to open display %s\n", displayname);
		exit(1);
	}
	scr = ScreenOfDisplay(dsp, screen);
	cmap = DefaultColormapOfScreen(scr);
	root = VirtualRootWindowOfScreen(scr);
	xswa.override_redirect = True;
	xswa.background_pixel = BlackPixel(dsp, screen);
	xswa.event_mask = (KeyPressMask |
			   ButtonPressMask |
			   VisibilityChangeMask);

	/* Constrain wide and tall to maximum sizes specified at top of
	   this file; this ensures that buffers will not be overshot,
	   but also allows VWWIDE and VWTALL to be used to override the
	   automatically detected dimensions, which are really virtual
	   (and hence may be larger than the physical screen presents)
	*/
	realxdim = WidthOfScreen(scr);
	realydim = HeightOfScreen(scr);
	if ((vwpanel.dim.x = realxdim) > VWWIDE) vwpanel.dim.x = VWWIDE;
	if ((vwpanel.dim.y = realydim) > VWTALL) vwpanel.dim.y = VWTALL;

	/* Make a window over the WHOLE screen anyway */
	win = XCreateWindow(dsp,
			    root,
			    0,
			    0,
			    realxdim,
			    realydim,
			    0,
			    CopyFromParent,
			    InputOutput,
			    CopyFromParent,
			    (CWOverrideRedirect |
			     CWBackPixel |
			     CWEventMask),
			    &xswa);
	XMapWindow(dsp, win);
	XRaiseWindow(dsp, win);

	xgcv.foreground = WhitePixel(dsp, screen);
	xgcv.background = BlackPixel(dsp, screen);
	my_gc = XCreateGC(dsp, win, (GCForeground | GCBackground), &xgcv);
	lockc = XCreateBitmapFromData(dsp, root, no_bits, 1, 1);
	lockm = XCreateBitmapFromData(dsp, root, no_bits, 1, 1);
	mycursor = XCreatePixmapCursor(dsp, lockc, lockm,
				       &nullcolor, &nullcolor, 0, 0);
	XFreePixmap(dsp, lockc);
	XFreePixmap(dsp, lockm);
	XGetScreenSaver(dsp, &sstimeout, &ssinterval,
			&ssblanking, &ssexposures);

	/* Ensure screen saver not active & then disable it */
	XResetScreenSaver(dsp);
	XSetScreenSaver(dsp, 0, 0, 0, 0);

	/* Select best panel depth available */
	{
		register int *depths;
		int numdepths = 0;
		int bestdepth = 8;
		depths = XListDepths(dsp, screen, &numdepths);
		while (numdepths--) {
			switch (depths[numdepths]) {
			case 8:
			case 16:
			case 24:
				if ((bestdepth < depths[numdepths]) ||
				    (bestdepth == 32)) {
					bestdepth = depths[numdepths];
				}
				break;
			case 32:
			
				if (bestdepth != 24) bestdepth = 32;
				break;
			}
		}

		vwpanel.depth = bestdepth;
		vwpanel.bytes = ((bestdepth / 8) *
				 vwpanel.dim.x *
				 vwpanel.dim.y);
	}

	/* Create X shared memory pixmap */
	vwximage = XShmCreateImage(dsp,
				   DefaultVisual(dsp, 0),
				   vwpanel.depth,
				   ZPixmap,
				   0,
				   &SHMInfo,
				   vwpanel.dim.x,
				   vwpanel.dim.y);

	/* Create the system V shared memory segment...
	   it is a newly created segment, accessible to anybody,
	   and set to self-destruct when nobody is using it
	*/
	SHMInfo.shmid = shmget(IPC_PRIVATE,
			       (vwximage->bytes_per_line *
				vwximage->height *
				1),
			       (IPC_CREAT | 0777));
	vwximage->data = SHMInfo.shmaddr = shmat(SHMInfo.shmid, 0, 0);
	XShmAttach(dsp, &SHMInfo);
	shmctl(SHMInfo.shmid, IPC_RMID, 0);

	/* Point buffer at the right spot */
	vwpanel.data.data = ((void *) vwximage->data);

	/* Clear the buffer */
	vpclear(&vwpanel);
}

inline void
vppaint(register vwbuf *panel)
{
	/* Anything to paint? */
	if (panel->dirty.min.x >= 0) {
#ifdef	PARTIALPAINT
printf("Partial paint update\n");
		XShmPutImage(dsp, win, my_gc,
			     vwximage,
			     panel->dirty.min.x,
			     panel->dirty.min.y,
			     panel->dirty.min.x,
			     panel->dirty.min.y,
			     (panel->dirty.max.x - panel->dirty.min.x),
			     (panel->dirty.max.y - panel->dirty.min.y),
			     0);
#else
		XShmPutImage(dsp, win, my_gc,
			     vwximage,
			     0,
			     0,
			     0,
			     0,
			     panel->dim.x,
			     panel->dim.y,
			     0);
#endif	PARTIALPAINT

		/* Make it so */
		XSync(dsp, 0);

		/* Not dirty anymore */
		panel->dirty.min.x = -1;
		panel->dirty.min.y = -1;
		panel->dirty.max.x = -1;
		panel->dirty.max.y = -1;
	}
}

void
vwpaint(register vwwall *wall)
{
	/* vwpaint(0) paints my panel no matter what;
	   vwpaint(wall) paints my panel only if it's in wall
	*/
	if (((!wall) && vwspace[IPROC]) ||
	    (wall == vwspace[IPROC])) {
		vppaint(&vwpanel);
	}
}

void
vwend(void)
{
	XSync(dsp, False);
	XClearWindow(dsp, win);

	XSetScreenSaver(dsp, sstimeout, ssinterval,
			ssblanking, ssexposures);
	XFlush(dsp);
	XCloseDisplay(dsp);
}


static int
xbadness(Display * d)
{
	exit(1);
}

void
vpbeginMouse(void)
{
	/* Grab mouse and keyboard */
	XGrabPointer(dsp,
		     win,
		     True,
		     (ButtonPressMask |
		      ButtonReleaseMask |
		      EnterWindowMask |
		      LeaveWindowMask |
		      PointerMotionMask |
		      PointerMotionHintMask |
		      Button1MotionMask |
		      Button2MotionMask |
		      Button3MotionMask |
		      Button4MotionMask |
		      Button5MotionMask |
		      ButtonMotionMask |
		      KeymapStateMask),
		     GrabModeAsync,
		     GrabModeAsync,
		     None,
		     mycursor,
		     CurrentTime);

#ifdef	NOTNOW
	XGrabKeyboard(dsp,
		      win,
		      True,
		      GrabModeAsync,
		      GrabModeAsync,
		      CurrentTime);
#endif

	/* Set dummy abort handler */
	XSetIOErrorHandler(xbadness);
}

static	vwinfoMouse prevMouse = { { 0.0, 0.0 }, 0, 0, 0 };

vwinfoMouse
vpreadMouse(register vwtime until)
{
	vwinfoMouse retval;

	do {
		Window rootret, childret;
		int rootx, rooty, winx, winy;
		unsigned int maskret;

		rootx = rooty = 0;
		XQueryPointer(dsp,
			      win,
			      &rootret,
			      &childret,
			      &rootx,
			      &rooty,
			      &winx,
			      &winy,
			      &maskret);

		retval.pos.x = (rootx *
				(1.0 / realxdim));
		retval.pos.y = (rooty *
				(1.0 / realydim));
		retval.left = ((maskret & Button1Mask) != 0);
		retval.center = ((maskret & Button2Mask) != 0);
		retval.right = ((maskret & Button3Mask) != 0);
	} while ((until != 0.0) &&
		 (vwnow() < until) &&
		 (prevMouse.pos.x == retval.pos.x) &&
		 (prevMouse.pos.y == retval.pos.y) &&
		 (prevMouse.left == retval.left) &&
		 (prevMouse.center == retval.center) &&
		 (prevMouse.right == retval.right));

	prevMouse = retval;
	return(retval);
}

void
vpendMouse(void)
{
	XUngrabPointer(dsp, CurrentTime);

#ifdef	NOTNOW
	XUngrabKeyboard(dsp, CurrentTime);
#endif	NOTNOW
}

void
vwbeginMouse(void)
{
	if (IPROC == CPROC)
		vpbeginMouse();
}

vwinfoMouse
vwreadMouse(register vwtime until)
{
	vwinfoMouse t;
	if (IPROC == CPROC) {
		t = vpreadMouse(until);
		p_bcastPutz(&t, sizeof(t));
	} else {
		p_bcastGetz(&t, sizeof(t));
	}
	return(t);
}

void
vwendMouse(void)
{
	if (IPROC == CPROC)
		vpendMouse();
}
