/*	control.c

	AFAPI control function definitions and data for WAPERS.

	July 1997 by Hank Dietz
*/
#if	NPROC == 1
#include "../ONE/one.c"
#endif
#if	NPROC > 1


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>


#include "intern.h"


int iproc = -1;			/* PE number (0..NPROC-1) */
int cproc = -1;			/* Console PE number */

unsigned short p_modeport;	/* Mode port address */
unsigned char p_outprev = 0x8;	/* Previous barrier bit */
unsigned int p_outbit = 0;	/* Single bit of data to output */

static int p_init_state = 0;	/* 0=never, 1=previously, 2=now */
static char *lockfile = NULL;



int
processControl(char *fname)
{
	FILE *theFile;
	struct stat stat_buf;

	/* The file should not exist, so stat() should fail, otherwise abort */
	if (!stat(fname, &stat_buf)) return(-1);

	if ((theFile = fopen(fname, "w")) == NULL) return(1);

	fprintf(theFile, "%10d\n", getpid());

	if (fclose(theFile)) return(5);
	lockfile = fname;
	return(0); /* sucess */
}

/*	p_init...

	This routine is called to check-in with the WAPERS hardware.
	It must be called before any other p_ or s_ routines.
*/

int
p_init(void)
{
	/* Initialize the WAPERS variables and hardware....  */
	register char *p;


	/* Step 0:  Have we done this before? */
	switch (p_init_state) {
	default:
		fprintf(stderr,
"AFAPI PE:  in p_init(), p_init_state=%d can't happen?\n",
			p_init_state);
		return(1);
	case 2:
		/* Call to init when we're already init?
		   We'll allow it, but do a p_confirm().
		*/
		goto step4;
	case 1:
		/* Yup, just need a partial check-in */
		goto step3;
	case 0:
		break;
	}


	/* Step 1:  Examine environment variables */
	p = getenv("AFAPI");
	if ((p == NULL) || strcmp(AFAPI, p)) {
		fprintf(stderr,
"AFAPI PE:  need %s hardware (set AFAPI environment variable)\n",
			AFAPI);
		return(2);
	}

	p = getenv("AFAPI_VERSION");
	if (p == NULL) p = "not given";
	if (strcmp(AFAPI_VERSION, p)) {
		fprintf(stderr,
"AFAPI PE:  AFAPI_VERSION of code is %s, environment is %s\n",
			AFAPI_VERSION, p);
		return(3);
	}

	p = getenv("AFAPI_NPROC");
	if ((p != NULL) && (NPROC != atoi(p))) {
		fprintf(stderr,
"AFAPI PE:  compiled for AFAPI_NPROC=%d, environment has %d\n",
			NPROC, atoi(p));
		return(4);
	}

	p = getenv("AFAPI_IPROC");
	iproc = ((p == NULL) ? -1 : atoi(p));
	if ((iproc < 0) || (iproc >= NPROC)) {
		fprintf(stderr,
"AFAPI PE:  AFAPI_IPROC is not a legal value (from 0 to %d)\n",
			(NPROC - 1));
		return(5);
	}

	p = getenv("AFAPI_CPROC");
	cproc = ((p == NULL) ? -1 : atoi(p));
	if ((cproc < 0) || (cproc >= NPROC)) {
		fprintf(stderr,
"AFAPI PE:  AFAPI_CPROC is not a legal value (from 0 to %d)\n",
			(NPROC - 1));
		return(6);
	}

	p = getenv("AFAPI_PORTBASE");
	if (p != NULL) {
		/* Set the port address variables */
		p_modeport = ((unsigned short) strtol(p, NULL, 0)) + 2;

#ifndef P_NON_STANDARD_PORT
		/* Is this one of the three standard parallel port
		   addresses?  if not, we shouldn't trust it....  Of
		   course, there is nothing inherently wrong with
		   having a parallel port at another address, so this
		   is code an informed user can change if appropriate.
		*/
		if ((p_modeport != (0x278 + 2)) &&
		    (p_modeport != (0x378 + 2)) &&
		    (p_modeport != (0x3bc + 2))) {
			fprintf(stderr,
"AFAPI PE%d:  AFAPI_PORTBASE=%d is not a standard parallel port\n",
				IPROC, (p_modeport - 2));
			return(7);
		}
#endif /* P_NON_STANDARD_PORT */
	} else {
		fprintf(stderr,
"AFAPI PE%d:  AFAPI_PORTBASE (parallel port address) was not set\n",
			IPROC);
		return(8);
	}


	/* Step 2:  Get permission to access the port.

	   This can use either of:

		ioperm(p_modeport, 1, 1)
		iopl(3)

	   The ioperm() call is safer, but yields slower port
	   operations due to I/O protection checks and is
	   Linux-specific.  Note that WAPERS only uses the mode port
	   -- the other two port registers, which are the only ones
	   used for TTL_PAPERS user code, are unused by WAPERS.
	*/
#ifdef P_USE_IOPL
        if (iopl(3)) {
#else
	if (ioperm(p_modeport, 1, 1)) {
#endif
		fprintf(stderr,
"AFAPI PE%d:  can't access parallel port (not priv)\n",
			IPROC);
		return(9);
	}


	/* Step 3:  Wake-up the connection and do an initial barrier.

	   This code *assumes* that the WAPERS "hardware" has been set
	   in a particular state that corresponds to having just
	   synchronized on barrier bit 0x8 with a 0 (inverted sense)
	   data bit.  This is a rather frail check, but it does have
	   the nice property that it should fail immediately if the
	   mode port is not really an open-collector TTL output....
	*/
step3:
	/* begin the meta-OS check-in */
	p = getenv("AFAPI_LOCKFILE");
	if (p == NULL) {
		fprintf(stderr,
"AFAPI PE%d:  AFAPI_LOCKFILE was not set\n",
			IPROC);
		return(8);
	} else if (processControl(p)) {
		fprintf(stderr,
"AFAPI PE%d:  A problem was encountered with the lockfile %s\n",
			IPROC, p);
		return(10);
	}

	/* begin the hardware check-in */
	if (((p__inb(p_modeport) ^
	      (0x8 | 0x2 | 0x1)) & (0x8 | 0x1)) != 0x8) {
		fprintf(stderr,
"AFAPI PE%d:  previous program did not check out or port not open-collector\n",
			IPROC);
		return(10);
	}

	/* Now, do an ordinary barrier on the 0x4 bit... */
	p_wait();


	/* Step 4:  Do a p_confirm(). */
step4:
	if (p_confirm() != 0) {
		fprintf(stderr,
"AFAPI PE%d:  p_confirm() suggests faulty hardware or connection\n",
			IPROC);
		return(11);
	}

	/* Step 5: Check that everyone agrees on the value of CPROC */
	if (p_any(p_reduceOr8u(CPROC) != CPROC)) {
		fprintf(stderr,
"AFAPI PE%d:  The CPROC environment variable is not consistent (=%d).\n",
			IPROC, CPROC);
		return(12);
	}


	/* Step 6:  We're in! */
	p_init_state = 2;
	return(0);
}

/*	p_exit()...

	Simply check-out with WAPERS.  Although most AFAPI programs
	will call this only just before exit(), it is also possible to
	check-out if your program will not need WAPERS for a while,
	and then check-in again using p_init().  However, check-in
	overhead is relatively high.
*/

void
p_exit(void)
{
	/* Step 0:  Well, where were we...? */
	switch (p_init_state) {
	default:
		fprintf(stderr,
"AFAPI PE:  in p_exit(), p_init_state=%d can't happen?\n",
			p_init_state);
		exit(1);
	case 0:
	case 1:
		/* We aren't even checked in...  so maybe the luser
		   wants us to confirm that we are really checked
		   out?  Paranoia?  Well, ok.  Why argue?
		*/
		goto step3;
	case 2:
		break;
	}


	/* Step 1:  Set the data bit to 0... */
	p_wait();
	p_outbit = 0;
	p_wait();

	/* Step 2:  Did everybody just sync on 0x8?  If not, do some
	   more syncs to get us there.
	*/
	while (p_outprev != 0x8) p_wait();

step3:
	/* Now remove the lockfile... this breaks the ability to do multiple
	   p_init/p_exit pairs in one program execution.  Oh well.
	 */
	if (lockfile == NULL) {
		fprintf(stderr,
"AFAPI PE%d:  in p_exit(), the lockfile variable was NULL.\n",
			IPROC);
		exit(1);
	} else if (remove(lockfile)) {
		fprintf(stderr,
"AFAPI PE%d:  in p_exit(), could not remove the lockfile '%s'.\n",
			IPROC, lockfile);
		exit(1);
	}

	/* Step 4:  We're out! */
	p_init_state = 1;
}

/*	p_confirm()...

	Attempt to confirm the previous operations.  In other words,
	if the hardware appears to have malfunctioned, return an error
	code; otherwise, return 0.  The idea is to allow the user to
	safely assume that correct operation has resumed by the time
	this code returns...  however, the electrically questionable
	wired-AND implementation of WAPERS makes this rather frail.
	If you are doing high-reliability stuff, DO NOT USE WAPERS!
*/

int
p_confirm(void)
{
	register int errs = -1;
	register int i, j;

again:
	if (++errs > (NPROC * NPROC * NPROC)) {
		fprintf(stderr,
"AFAPI PE%d:  in p_confirm(), unrecoverable hardware fault\n",
			IPROC);
		exit(1);
	}

	if (!p_all(1)) goto again;

	for (i=0; i<NPROC; ++i) {
		if (!p_all(1)) {
			goto again;
		}

		for (j=0; j<=i; ++j) {
			if (p_all(i != IPROC)) {
				goto again;
			}
			if (!p_all(1)) {
				goto again;
			}
		}
	}

	/* Return 1 for error, 0 otherwise */
	return(p_any(errs));
}
#endif /* if NPROC > 1 */
