/*	vote.c

	AFAPI voting and scheduling function definitions for
	WAPERS.

	July 1997 by H. Dietz
*/
#if	NPROC > 1


#include "intern.h"

/*	p_first()...

	Return the IPROC of the first lowest-numbered PE with a yes
	vote.  A binary search is done to find the first f != 0.
*/
int
p_first(register p_1u f)
{
	register int hi = NPROC - 1, lo = 0, pivot;

	do {
		pivot = (hi + lo) >> 1;
		if (p__any((IPROC >= lo) && (IPROC <= pivot) && f)) {
			hi = pivot;
		} else { /* none found in the lower half */
			lo = pivot + 1;
		}
	} while (hi > lo);
	if ((lo < NPROC - 1) || p__any(f)) { /* Was there at least one? */
		return(lo);
	} else {
		return(NPROC);
	}
}

/*	p_count()...

	Count how many voted yes.  This uses a somewhat sneaky scheme
	that essentially checks for nobody and everybody.  If those
	fail, it loops counting, but stops before the last bit to try
	to deduce the value before resorting to sampling it.
*/

int
p_count(register p_1u f)
{
	/* First check for nobody */
	if (!p__any(f)) return(0);

	/* Then check for everybody */
	if (p__all(f)) return(NPROC);

	/* Count them one by one */
	{
		register int i;
		register int count = 0;

		for (i=0; i<(NPROC-1); ++i) {
			count += p__all((IPROC != i) || f);
		}

		/* Try to deduce the state of the last f */
		if (count == (NPROC - 1)) return(NPROC - 1);
		if (count == 0) return(1);

		/* No luck.  Check the last one */
		count += p__all((IPROC != (NPROC - 1)) || f);
		return(count);
	}
}

/*	p_quantify()...

	Quantify the number of PEs voting yes; the return is 0, 1, 2
	(really >1), or NPROC.  If NPROC is 3 or less, p_quantify
	returns an exact count.  As of the 9807xx version of AFAPI,
	p_quantify() will return identical values regardless of which
	implementation of AFAPI is used.
*/

int
p_quantify(register p_1u f)
{
	/* First check for nobody */
	if (!p__any(f)) return(0);

	/* Then check for everybody */
	if (p__all(f)) return(NPROC);

	/* Then find one, and see if there is another */
	{
		register int i;

		for (i=0; i<(NPROC - 1); ++i) {
			if (p__all((IPROC != i) || f)) {
				return(p__any((IPROC > i) && f) ? 2 : 1);
			}
		}

		/* Didn't find any, so the last one was it */
		return(1);
	}

	/* This comment is not reachable */
}

/*	p_vote()...

	Builds a vote barrier bit vector using one p__waitbar() to
	determine an "upper bound" mask and then a sequence of Log2
	NPROC p__waitbar(), each of which determines who matched 1 bit
	of everybody's IPROC.
*/

inline p_bar
p_vote(register int pe)
{
	register p_bar bit = 1;
	register p_bar votemask = p__waitbar((pe >= 0) && (pe < NPROC));

	do {
		if (IPROC & bit) {
			votemask &= p__waitbar(pe & bit);
		} else {
			votemask &= ~p__waitbar(pe & bit);
		}
	} while ((bit <<= 1) & (NPROC-1));

	return(votemask);
}

/*	p_voteCount()...

	Uses p_vote, then count the bits.
*/

int
p_voteCount(register int pe)
{
	register int i;
	register int count = 0;
	register p_bar votemask = p_vote(pe);

	for (i=0; i<NPROC; ++i) {
		count += (votemask & 1);
		votemask >>= 1;
	}

	return(count);
}

/*	p_match()...
*/

#define	P_DECLARE(suffix) \
inline p_bar \
p_match##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register p_bar mask = 1; \
	register p_bar matchmask = (1 << IPROC); \
 \
	do { \
		if (IPROC == who) { \
			/* Our turn to send and store data */ \
			p_bcastPut##suffix(datum); \
		} else { \
			/* We just get and compare data... */ \
			if (datum == p_bcastGet##suffix()) { \
				matchmask |= mask; \
			} \
		} \
		mask <<= 1; \
	} while (++who < NPROC); \
 \
	return(matchmask); \
}

P_BETA

#undef	P_DECLARE

/*	p_matchCount()...

	Very much like p_match(), but doesn't bother with the mask;
	just counting is quicker.
*/

#define	P_DECLARE(suffix) \
int \
p_matchCount##suffix(register p_##suffix datum) \
{ \
	register int who = 0; \
	register int count = 1; /* We match ourself */ \
 \
	do { \
		if (IPROC == who) { \
			/* Our turn to send data */ \
			p_bcastPut##suffix(datum); \
		} else { \
			/* We just get and compare data... */ \
			count += (datum == p_bcastGet##suffix()); \
		} \
	} while (++who < NPROC); \
 \
	return(count); \
}

P_BETA

#undef	P_DECLARE

/*	p_rank()...
*/

#define	P_DECLARE(suffix) \
int \
p_rank##suffix(register p_##suffix datum) \
{ \
	register p_##suffix d; \
	register int who = 0; \
	register int rank = 0; \
 \
	do { \
		if (IPROC == who) { \
			/* Output our value */ \
			p_bcastPut##suffix(d = datum); \
		} else { \
			/* Increment our rank iff his value is bigger \
			   or his value equals our value and he has a \
			   lower value of IPROC. \
			*/ \
			register p_##suffix t = p_bcastGet##suffix(); \
 \
			rank += ((t < datum) || \
				 ((t == datum) && (who < IPROC))); \
		} \
	} while (++who < NPROC); \
 \
	return(rank); \
}

P_ALPHA

#undef	P_DECLARE

/*	p_ranked()...

	Well, rank was such a handy operation on the MasPar...  it
	never occurred to us that it is backward.  Basically, if you
	want to sort using putGet, you don't care where your data
	goes, but simply where the datum you should get will have to
	come from.  The ranked operation generalizes this so that each
	PE can get the current location of the value that would be
	ranked as the second argument specifies.  This is currently
	done by a dumb gather and (rank-truncated) bubble sort.
*/

#define	P_DECLARE(suffix) \
int \
p_ranked##suffix(register p_##suffix datum, \
register int rank) \
{ \
	p_##suffix buf[NPROC]; \
	int index[NPROC]; \
	register int i, j; \
 \
	p_gather##suffix(&(buf[0]), datum); \
	if (rank >= NPROC) return(NPROC); \
	for (i=0; i<NPROC; ++i) index[i] = i; \
	for (i=0; i<=rank; ++i) { \
		for (j=i+1; j<NPROC; ++j) { \
			if (buf[index[i]] > buf[index[j]]) { \
				register int t = index[i]; \
				index[i] = index[j]; \
				index[j] = t; \
			} \
		} \
	} \
 \
	return(index[rank]); \
}

P_ALPHA

#undef	P_DECLARE
#endif /* if NPROC > 1 */
