/*	afreset.c

	AFAPI reset of TTL_PAPERS & WAPERS hardware to the "off" state.

	Under normal circumstances, this helper program is not needed.
	However, it is the only way to "check out" without first
	checking in, which is important if a PE process terminates
	without first checking out....

	1996 by Hank Dietz

	Added 980217 by T.I.Mattox
	afreset now reads the AFAPI_LOCKFILE env variable for the name of
	a file that contains a the PID number in ASCII of a process that
	should be killed before attempting to reset the parallel port.
	That process is sent a SIGINT, SIGTERM, and SIGKILL in sequence with
	pauses to allow the process a chance to shut down gracefully.
	The file is removed if afreset is sucessful in killing the process.
	
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>

#include <AFAPI/version.h>


/* TTL_PAPERS variables */
unsigned short p_outport;	/* Parallel port output address */
unsigned short p_inport;	/* Parallel port input address */
unsigned char p_outlast;	/* Last port output value */

/* WAPERS variables */
unsigned short p_modeport;      /* Mode port address */
unsigned char p_outprev = 0x8;  /* Previous barrier bit */
unsigned int p_outbit = 0;      /* Single bit of data to output */

/*	ioperm() and iopl()

	Function prototypes included here since it's location in the standard
	header file tree has moved from version to version of Linux:
		Under Libc5 these are in <unistd.h>
		Under Glibc1 these are left unspecified!
		Under Glibc2 these are in <sys/io.h>  (and <sys/perm,h>)
*/
extern int
ioperm(unsigned long from, unsigned long num, int turn_on);

#ifdef P_USE_IOPL
extern int
iopl(int level);
#endif


extern inline
unsigned int
p__inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}


extern inline
void
p__outb(unsigned char value,
unsigned short port)
{
__asm__ __volatile__ ("outb %b0,%w1"
		:/* no outputs */
		:"a" (value), "d" (port));
}

int
registerOurPID(const char *fname)
{
	FILE *theFile;

	if ((theFile = fopen(fname, "w")) == NULL) {
		fprintf(stderr,"afreset: Could not open/create '%s'.\n",fname);
		return(1);
	}
	fprintf(theFile, "%10d\n", getpid());
	if (fclose(theFile)) {
		fprintf(stderr,"afreset: Strange, fclose(%s) failed.\n",fname);
		return(5);
	}
	return(0);
}

/* afreset will avoid killing processes with PID less than SPECIAL_PID */
/* Adjust it to encompas the daemons started at boot time on your system. */
#define	SPECIAL_PID	100

int
processControl(const char *fname)
{
	FILE *theFile;
	struct stat stat_buf;
	int thePID, i;
	int sigList[] = { SIGINT, SIGTERM, SIGKILL, 0 };
	int waitList[] = { 1000000, 500000, 0, 0 }; /* in microseconds */

	if (stat(fname, &stat_buf)) {
		/* this is considered OK */
		/* fprintf(stderr, "afreset: '%s' does not exist.\n", fname); */
		return(registerOurPID(fname));
	}
	if (!S_ISREG(stat_buf.st_mode)) {
		fprintf(stderr,"afreset: '%s' is not a regular file.\n",fname);
		return(3);
	}
	if ((theFile = fopen(fname, "r")) == NULL) {
		fprintf(stderr, "afreset: Could not open '%s'.\n", fname);
		return(4);
	}

	if (fscanf(theFile, "%d", &thePID) != 1) {
		fprintf(stderr,"afreset: '%s' did not contain a PID.\n",fname);
		fclose(theFile);
		return(registerOurPID(fname));
	}
	
	if (fclose(theFile)) {
		fprintf(stderr,"afreset: Strange, fclose(%s) failed.\n",fname);
		return(5);
	}

	if ((thePID <= SPECIAL_PID) && (thePID >= -SPECIAL_PID)) {
		fprintf(stderr, "afreset: The PID %d is special.\n", thePID);
		return(6);
	}

	for (i = 0; sigList[i]; ++i) {
		if (kill(thePID, sigList[i]) >= 0) {
			usleep(waitList[i]);
			continue;
		} else if (errno == ESRCH) {
/*
			fprintf(stderr,
				"afreset: The PID %d is now gone (%d).\n",
				thePID, i);
*/
			break;
		}
		switch (errno) {
			case EPERM: fprintf(stderr,
	"afreset: Permission denied on PID %d.\n", thePID);
				return(7);
			default: fprintf(stderr,
	"afreset: SIGTERM failed on PID %d.\n", thePID);
				return(8);
		}
	}

	return(registerOurPID(fname));
}

void
reset_TTL_PAPERS()
{

	/* Turn off the parallel interrupt on TTL_PAPERS */
	p__outb(0x01, p_modeport);

	/* Step 2:  Set P_S1, P_LG, and all data bits so that the same
	   thing we already see on the input port would be actively
	   generated, except, of course, for the fact that we also
	   turn on the green and red LEDs (we are OS code).
	*/
	p_outlast = (0x80 | 0x20 | 0x10 | 0x0f);
	p__outb(p_outlast, p_outport);

	while ((p__inb(p_inport) & 0xf8) != 0) ;

	/* Step 3:  Turn off the LEDs and lock the strobes */
	p_outlast = (0x0f);
	p__outb(p_outlast, p_outport);
}

void
reset_WAPERS()
{
	register int tries = 0;

	/* Step 2:  Put us in the off state */
	p__outb(((0x8 | 0x2) ^ (0x8 | 0x2 | 0x1)), p_modeport);

	/* Step 3:  Are we there yet? */
	while (((p__inb(p_modeport) ^
		 (0x8 | 0x2 | 0x1)) &
		 (0x8 | 0x4 | 0x2 | 0x1)) != (0x8 | 0x2)) {
		if ((tries++ & ((1024 * 1024) - 1)) == ((1024 * 1024) - 1)) {
			register int pe = -1;
			register char *p = getenv("AFAPI_IPROC");
			if (p) pe = atoi(p);
			fprintf(stderr,
"AFAPI reset:  after %d polls, PE%d still waiting for hardware reset\n",
				tries,
				pe);
			fprintf(stderr,
"AFAPI reset:  are you sure PE%d has a compliant SPP port at 0x%x?\n",
				pe,
				(p_modeport - 2));
		}
	}
}

int
main(register int argc,
register char **argv)
{
	register char *p, *lockfile;
	int flavor = 0, fileErr = 0;

	/* Step 0:  Examine environment variables */
	p = getenv("AFAPI");
	if (p == NULL) {
		fprintf(stderr,
"AFAPI reset:  The AFAPI mechanism environment variable has not been set.\n");
		exit(1);
	}

	if (strcmp("TTL_PAPERS", p) == 0) {
		flavor = 1;
	} else if (strcmp("WAPERS", p) == 0) {
		flavor = 2;
	} else {
		fprintf(stderr,
"AFAPI reset:  %s mechanism is not supported by this version %s.\n",
		p, AFAPI_VERSION);
		exit(2);
	}

	p = getenv("AFAPI_VERSION");
	if (p == NULL) p = "not given";
	if (strcmp(AFAPI_VERSION, p)) {
		fprintf(stderr,
	"AFAPI Warning:  AFAPI_VERSION of code is %s, environment is %s\n",
			AFAPI_VERSION, p);
	}

	p = getenv("AFAPI_PORTBASE");
	if (p != NULL) {
		/* Set the port address variables */
		p_outport = (unsigned short) strtol(p, NULL, 0);
		p_inport = p_outport + 1;
		p_modeport = p_outport + 2;

#ifndef P_NON_STANDARD_PORT
		/* Is this one of the three standard parallel port
		   addresses?  if not, we shouldn't trust it....  Of
		   course, there is nothing inherently wrong with
		   having a parallel port at another address, so this
		   is code an informed user can change if appropriate.
		*/
		if ((p_outport != 0x278) &&
		    (p_outport != 0x378) &&
		    (p_outport != 0x3bc)) {
			fprintf(stderr,
"AFAPI reset:  AFAPI_PORTBASE=%d is not a standard parallel port\n",
				p_outport);
			exit(1);
		}
#endif /* P_NON_STANDARD_PORT */
	} else {
		fprintf(stderr,
"AFAPI reset:  AFAPI_PORTBASE (parallel port address) was not set\n");
		exit(1);
	}


	/* Step 1:  Get permission to access the port.

	   This can use any of:

		ioperm(p_outport, 3, 1)
		iopl(3)

	   The ioperm() call is safer, but yields slower port
	   operations due to I/O protection checks and is
	   Linux-specific.
	*/
#ifdef P_USE_IOPL
        if (iopl(3)) {
#else
        if (ioperm(p_outport, 3, 1)) {
#endif
		fprintf(stderr,
			"AFAPI PE:  can't access parallel port\n");
		exit(1);
	}

	lockfile = getenv("AFAPI_LOCKFILE");
	if (lockfile != NULL) fileErr = processControl(lockfile);
	else fprintf(stderr,"afreset: Warning, AFAPI_LOCKFILE was not set.\n");

	switch (flavor) {
		case 1: reset_TTL_PAPERS();
			break;
		case 2: reset_WAPERS();
			break;
		default: break;
	}

	/* we are done, remove the AFAPI_LOCKFILE */
	if ((lockfile != NULL) && (!fileErr) && remove(lockfile)) {
		fprintf(stderr, "afreset: Could not remove '%s'.\n", lockfile);
	}
	return(0);
}
