/*	portinfo.c

	This routine checks compatibility of a parallel port with
	WAPERS.  It doesn't do anything magical, and I suppose that
	there might be a port design out there somewhere that will
	cause this routine to get the wrong answer, but at least this
	makes a reasonable attempt at checking:

	* What port address are you using?
	* Does the control port have TTL open-collector outputs?
	* Does the data port have tri-state ability?

	This check is done using a very simple test plug in which only
	pins 1 and 2 are connected...  easily done using a solder
	bridge on a DB25 connector.

	July 1997 by Hank Dietz
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

/*	iopl()

	Function prototype included here since it's location in the standard
	header file tree has moved from version to version of Linux:
		Under Libc5 these are in <unistd.h>
		Under Glibc1 these are left unspecified!
		Under Glibc2 these are in <sys/io.h>  (and <sys/perm,h>)
*/
extern int
iopl(int level);

double cin, cout, din, dout;


extern inline
unsigned int
p__inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}


extern inline
void
p__outb(unsigned char value,
unsigned short port)
{
__asm__ __volatile__ ("outb %b0,%w1"
		:/* no outputs */
		:"a" (value), "d" (port));
}

extern void timeall(register int a, register int cantri);

int ports[] = { 0x278, 0x378, 0x3bc };
int tristate[] = { 0, 0, 0 };

#define	forports(i, p) \
	for (p=ports[i=0]; i<3; p=ports[++i]) if (p)

int
yes()
{
	int c;

	printf("Type n<ret> for no, <ret> for yes: ");
	c = getchar();
	if (c != '\n') {
		while (getchar() != '\n');
	}
	return(c != 'n');
}


int
main(register int argc,
register char **argv)
{
	register int a, i, j, k;
	register char *p;

	/* Step 0: get I/O access permission */
	if (iopl(3)) {
		fprintf(stderr,
			"%s:  must be priv to access parallel ports\n",
			argv[0]);
		exit(1);
	}

	/* Step 1: set all the ports to a "safe" state.

	   Because pins 1 and 2 are tied together on the test jig, and
	   both ports may be actively driven, we immediately set all
	   the port values to something safe...  just in case they
	   have already plugged-in the test jig.  First, set the
	   control port to output all high.  Then set the data port to
	   output high.  All high should be harmless....
	*/
	forports(i, a) {
		p__outb(0x04, (a + 2));
		p__outb(0xff, a);
	}

	/* Step 2: scan for ports.

	   A port is present where toggling an output bit causes the
	   same value to be read back from the feedback input register.
	*/
	k = 0;
	forports(i, a) {
		p__outb(0x7f, a);
		j = p__inb(a);
		p__outb(0xff, a);
		j = ((j << 8) | p__inb(a));

		switch (j) {
		case 0x7fff:
			/* We have a parallel port, is it ECP? */
			p = "SPP";
			j = p__inb(a + 0x402);
			if ((j & 3) == 1) {
				/* ECR says empty + not full ECP FIFO;
				   now set the ECP port into PS/2 mode.
				   This is actually the standard ECP
				   detection procedure....
				*/
				if (p__inb(a + 0x402) == 0x35) {
					p = "ECP";
				} else {
					p = "ECP?";
				}
			} else if (a != 0x3bc) {
				/* If not at 0x3bc and not ECP, is it EPP?
				   First clear the EPP status, then write
				   to one of the extended ports, read it
				   back, and clear the status
				*/
				p__outb(0x0, (a + 1));
				p__outb(0x1, (a + 1));
				p__inb(a + 1);
				p__outb(0x7f, (a + 3));
				if (p__inb(a + 3) == 0x7f) {
					p = "EPP";
					p__outb(0x0, (a + 1));
					p__outb(0x1, (a + 1));
					p__inb(a + 1);
					p__outb(0xff, a);
				}
			}

			/* Can we tri-state disable the data port? */
			p__outb((0x20 | 0x04), (a + 2));
			p__outb(0x7f, a);
			j = p__inb(a);
			p__outb(0xff, a);
			j ^= p__inb(a);
			p__outb(0x04, (a + 2));
			tristate[i] = !j;

			/* Let user know what we found */
printf("%s %s parallel port detected at 0x%x (%d)\n",
				p,
				(j ? "non-tri-state" : "tri-state"),
				a, a);
			++k;
			break;
		default:
			ports[i] = 0;
		}
	}
	if (k) {
printf("%d usable ports were detected.\n", k);
	} else {
printf("No usable parallel ports were detected.\n");
		exit(1);
	}
printf("\n");

	/* Step 3: tell user what to do.
	*/
printf("Up to this point, there should have been nothing connected to\n");
printf("the parallel port of this computer that you intend to use for\n");
printf("WAPERS.  Now you should connect a special test plug to the\n");
printf("desired port.  This test plug should connect pins 1 and 2; all\n");
printf("other pins should be unconnected.\n");
printf("\n");

	/* Step 4: does the user want to do the real test?
	*/
step4:
printf("It is not safe to use WAPERS without having checked the port\n");
printf("using the test plug, but port timing can be tested without it.\n");
printf("Simply timing all the ports can be useful in deciding which of\n");
printf("several alternative ports might perform best for WAPERS.\n");
printf("\n");
printf("Do you want to connect the test plug?\n");
	if (!yes()) {
		forports(i, a) {
			timeall(a, tristate[i]);
		}
		exit(1);
	}
printf("\n");
	do {
printf("Now connect the test plug.  Is it connected?\n");
	} while (!yes());
printf("\n");

	/* Step 5: scan for the test plug.
	*/
	forports(i, a) {
		/* First, see if the control port is open collector */
		p__outb(0xfe, a);
		j = p__inb(a + 2);
		p__outb(0xff, a);
		if ((j & 0x0f) == (0x04 | 0x01)) {
			/* Ok, control port is open collector */
printf("Test plug detected at 0x%x (%d):\n", a, a);
printf("Control port (0x%x) is open collector.\n", (a + 2));

			/* Can data port be tri-state disabled? */
			p__outb((0x20 | 0x04), (a + 2));
			p__outb(0xfe, a);
			j = p__inb(a + 2);
			p__outb(0xff, a);
			if ((j & 0x0f) == (0x04 | 0x01)) {
printf("Data port (0x%x) CANNOT be tri-state disabled.\n", a);
printf("WAPERS should work, but can only use 4-bit wired-AND.\n");
printf("\n");
printf("The power-on setup facility of your computer may allow you to\n");
printf("change port configuration so tri-state disabling will work:\n");
printf("PS/2 or ECP modes are most likely to work.  However, be sure\n");
printf("to run this test program again after changing the mode; some\n");
printf("port modes may not use open-collector drive for the control\n");
printf("port outputs, in which case WAPERS cannot be used at all.\n");
			} else {
printf("Data port (0x%x) can be tri-state disabled.\n", a);
printf("WAPERS can use both 4-bit wired-AND and 8-bit broadcast.\n");
			}
printf("\n");
printf("You should disconnect the test plug now.\n");
			timeall(a, tristate[i]);
			exit(0);
		}
	}

	/* Step 6: if test plug was connected, WAPERS will not work.
	*/
printf("Test plug not detected.\n");
printf("Either the test plug is not connected or your parallel port(s)\n");
printf("do not implement the control output lines using open-collector\n");
printf("TTL.  Are you sure the test plug is connected?\n");
	if (yes()) {
printf("\n");
printf("Sorry to hear that....\n");
printf("\n");
printf("You should disconnect the test plug now.\n");
printf("\n");
printf("You should not attempt to use WAPERS with this port as it is\n");
printf("configured -- if you do, it may damage your port hardware.\n");
printf("The power-on setup facility of your computer may allow you to\n");
printf("change the configuration so that it will work: generally, SPP,\n");
printf("PS/2, or ECP are the modes most likely to work.\n");
printf("\n");
printf("Alternatively, you might purchase a cheap parallel port card\n");
printf("that implements both the open-collector control port and the\n");
printf("tri-state disable of the data output port drivers.\n");
	} else {
printf("\n");
printf("Well, then...  that proves nothing....\n");
printf("\n");
		goto step4;
	}
	return(1);
}


extern void p_timerStart(void);
extern double p_timerElapsed(void);

#include <sys/time.h>

static struct timeval dpu_start, dpu_stop;


void
p_timerStart()
{
	gettimeofday(&dpu_start, 0);
}


double
p_timerElapsed()
{
	gettimeofday(&dpu_stop, 0);
	return(((1000000.0 * (dpu_stop.tv_sec - dpu_start.tv_sec)) +
		(dpu_stop.tv_usec - dpu_start.tv_usec) + 0.5) / 1000000.0);
}

void
timeall(register int a,
register int cantri)
{
	register int a2 = (a + 2);
	register int v = (0x20 | 0x04);
	register int i;

printf("\n");
printf("Timing control port 0x%x output speed:  ", a2);
	p_timerStart();
	for (i=0; i<100000; ++i) {
		p__outb(v, a2);
		p__outb(v, a2);
		p__outb(v, a2);
		p__outb(v, a2);
		p__outb(v, a2);
		p__outb(v, a2);
		p__outb(v, a2);
		p__outb(v, a2);
		p__outb(v, a2);
		p__outb(v, a2);
	}
printf("%gus\n", cout = p_timerElapsed());

printf("Timing control port 0x%x input speed:   ", a2);
	p_timerStart();
	for (i=0; i<100000; ++i) {
		p__inb(a2);
		p__inb(a2);
		p__inb(a2);
		p__inb(a2);
		p__inb(a2);
		p__inb(a2);
		p__inb(a2);
		p__inb(a2);
		p__inb(a2);
		p__inb(a2);
	}
printf("%gus\n", cin = p_timerElapsed());

	if (cantri) {
printf("Timing data port 0x%x output speed:     ", a);
		v = 0xff;
		p_timerStart();
		for (i=0; i<100000; ++i) {
			p__outb(v, a);
			p__outb(v, a);
			p__outb(v, a);
			p__outb(v, a);
			p__outb(v, a);
			p__outb(v, a);
			p__outb(v, a);
			p__outb(v, a);
			p__outb(v, a);
			p__outb(v, a);
		}
printf("%gus\n", dout = p_timerElapsed());

printf("Timing data port 0x%x input speed:      ", a);
		p_timerStart();
		for (i=0; i<100000; ++i) {
			p__inb(a);
			p__inb(a);
			p__inb(a);
			p__inb(a);
			p__inb(a);
			p__inb(a);
			p__inb(a);
			p__inb(a);
			p__inb(a);
			p__inb(a);
		}
printf("%gus\n", din = p_timerElapsed());
	}

printf("\n");
printf("Expected speed of some WAPERS operations on port 0x%x:\n", a);
printf("Barrier synchronization, p_wait() ................ %gus\n",
       (cout + cin));
printf("One bit AND, p_all() ............................. %gus\n",
       (cout + cin + cout + cin + ((cin > cout) ? cin : cout)));
printf("Bandwidth (wired-AND, without compression) ....... %d bits/s\n",
       ((int) (1000000.0 /
        (cout + cin + cout + cin + ((cin > cout) ? cin : cout)))));
printf("Bandwidth (wired-AND, compressed best case) ...... %d bits/s\n",
       ((int) (((64.0 / 7.0) * 1000000.0) /
        (cout + cin + cout + cin + ((cin > cout) ? cin : cout)))));
	if (cantri) {
printf("Bandwidth (tri-state bus, without compression) ... %d bits/s\n",
       ((int) (8000000.0 /
        (cout + cin + cout + cin + ((din > dout) ? din : dout)))));
	}
}
